"""Miscellaneous type operations and helpers for use during type checking.

NOTE: These must not be accessed from mypy.nodes or mypy.types to avoid import
      cycles. These must not be called from the semantic analysis main pass
      since these may assume that MROs are ready.
"""

from __future__ import annotations

import itertools
from collections.abc import Iterable, Sequence
from typing import Any, Callable, TypeVar, cast

from mypy.copytype import copy_type
from mypy.expandtype import expand_type, expand_type_by_instance
from mypy.maptype import map_instance_to_supertype
from mypy.nodes import (
    ARG_OPT,
    ARG_POS,
    ARG_STAR,
    ARG_STAR2,
    SYMBOL_FUNCBASE_TYPES,
    Decorator,
    Expression,
    FuncBase,
    FuncDef,
    FuncItem,
    OverloadedFuncDef,
    StrExpr,
    SymbolNode,
    TypeInfo,
    Var,
)
from mypy.state import state
from mypy.types import (
    AnyType,
    CallableType,
    ExtraAttrs,
    FormalArgument,
    FunctionLike,
    Instance,
    LiteralType,
    NoneType,
    NormalizedCallableType,
    Overloaded,
    Parameters,
    ParamSpecType,
    PartialType,
    ProperType,
    TupleType,
    Type,
    TypeAliasType,
    TypedDictType,
    TypeOfAny,
    TypeQuery,
    TypeType,
    TypeVarLikeType,
    TypeVarTupleType,
    TypeVarType,
    UninhabitedType,
    UnionType,
    UnpackType,
    flatten_nested_unions,
    get_proper_type,
    get_proper_types,
)
from mypy.typetraverser import TypeTraverserVisitor
from mypy.typevars import fill_typevars


def is_recursive_pair(s: Type, t: Type) -> bool:
    """Is this a pair of recursive types?

    There may be more cases, and we may be forced to use e.g. has_recursive_types()
    here, but this function is called in very hot code, so we try to keep it simple
    and return True only in cases we know may have problems.
    """
    if isinstance(s, TypeAliasType) and s.is_recursive:
        return (
            isinstance(get_proper_type(t), (Instance, UnionType))
            or isinstance(t, TypeAliasType)
            and t.is_recursive
            # Tuple types are special, they can cause an infinite recursion even if
            # the other type is not recursive, because of the tuple fallback that is
            # calculated "on the fly".
            or isinstance(get_proper_type(s), TupleType)
        )
    if isinstance(t, TypeAliasType) and t.is_recursive:
        return (
            isinstance(get_proper_type(s), (Instance, UnionType))
            or isinstance(s, TypeAliasType)
            and s.is_recursive
            # Same as above.
            or isinstance(get_proper_type(t), TupleType)
        )
    return False


def tuple_fallback(typ: TupleType) -> Instance:
    """Return fallback type for a tuple."""
    info = typ.partial_fallback.type
    if info.fullname != "builtins.tuple":
        return typ.partial_fallback
    items = []
    for item in typ.items:
        if isinstance(item, UnpackType):
            unpacked_type = get_proper_type(item.type)
            if isinstance(unpacked_type, TypeVarTupleType):
                unpacked_type = get_proper_type(unpacked_type.upper_bound)
            if (
                isinstance(unpacked_type, Instance)
                and unpacked_type.type.fullname == "builtins.tuple"
            ):
                items.append(unpacked_type.args[0])
            else:
                raise NotImplementedError
        else:
            items.append(item)
    return Instance(
        info,
        # Note: flattening recursive unions is dangerous, since it can fool recursive
        # types optimization in subtypes.py and go into infinite recursion.
        [make_simplified_union(items, handle_recursive=False)],
        extra_attrs=typ.partial_fallback.extra_attrs,
    )


def get_self_type(func: CallableType, default_self: Instance | TupleType) -> Type | None:
    if isinstance(get_proper_type(func.ret_type), UninhabitedType):
        return func.ret_type
    elif func.arg_types and func.arg_types[0] != default_self and func.arg_kinds[0] == ARG_POS:
        return func.arg_types[0]
    else:
        return None


def type_object_type(info: TypeInfo, named_type: Callable[[str], Instance]) -> ProperType:
    """Return the type of a type object.

    For a generic type G with type variables T and S the type is generally of form

      Callable[..., G[T, S]]

    where ... are argument types for the __init__/__new__ method (without the self
    argument). Also, the fallback type will be 'type' instead of 'function'.
    """

    # We take the type from whichever of __init__ and __new__ is first
    # in the MRO, preferring __init__ if there is a tie.
    init_method = info.get("__init__")
    new_method = info.get("__new__")
    if not init_method or not is_valid_constructor(init_method.node):
        # Must be an invalid class definition.
        return AnyType(TypeOfAny.from_error)
    # There *should* always be a __new__ method except the test stubs
    # lack it, so just copy init_method in that situation
    new_method = new_method or init_method
    if not is_valid_constructor(new_method.node):
        # Must be an invalid class definition.
        return AnyType(TypeOfAny.from_error)

    # The two is_valid_constructor() checks ensure this.
    assert isinstance(new_method.node, (SYMBOL_FUNCBASE_TYPES, Decorator))
    assert isinstance(init_method.node, (SYMBOL_FUNCBASE_TYPES, Decorator))

    init_index = info.mro.index(init_method.node.info)
    new_index = info.mro.index(new_method.node.info)

    fallback = info.metaclass_type or named_type("builtins.type")
    if init_index < new_index:
        method: FuncBase | Decorator = init_method.node
        is_new = False
    elif init_index > new_index:
        method = new_method.node
        is_new = True
    else:
        if init_method.node.info.fullname == "builtins.object":
            # Both are defined by object.  But if we've got a bogus
            # base class, we can't know for sure, so check for that.
            if info.fallback_to_any:
                # Construct a universal callable as the prototype.
                any_type = AnyType(TypeOfAny.special_form)
                sig = CallableType(
                    arg_types=[any_type, any_type],
                    arg_kinds=[ARG_STAR, ARG_STAR2],
                    arg_names=["_args", "_kwds"],
                    ret_type=any_type,
                    is_bound=True,
                    fallback=named_type("builtins.function"),
                )
                return class_callable(sig, info, fallback, None, is_new=False)

        # Otherwise prefer __init__ in a tie. It isn't clear that this
        # is the right thing, but __new__ caused problems with
        # typeshed (#5647).
        method = init_method.node
        is_new = False
    # Construct callable type based on signature of __init__. Adjust
    # return type and insert type arguments.
    if isinstance(method, FuncBase):
        t = function_type(method, fallback)
    else:
        assert isinstance(method.type, ProperType)
        assert isinstance(method.type, FunctionLike)  # is_valid_constructor() ensures this
        t = method.type
    return type_object_type_from_function(t, info, method.info, fallback, is_new)


def is_valid_constructor(n: SymbolNode | None) -> bool:
    """Does this node represents a valid constructor method?

    This includes normal functions, overloaded functions, and decorators
    that return a callable type.
    """
    if isinstance(n, SYMBOL_FUNCBASE_TYPES):
        return True
    if isinstance(n, Decorator):
        return isinstance(get_proper_type(n.type), FunctionLike)
    return False


def type_object_type_from_function(
    signature: FunctionLike, info: TypeInfo, def_info: TypeInfo, fallback: Instance, is_new: bool
) -> FunctionLike:
    # We first need to record all non-trivial (explicit) self types in __init__,
    # since they will not be available after we bind them. Note, we use explicit
    # self-types only in the defining class, similar to __new__ (but not exactly the same,
    # see comment in class_callable below). This is mostly useful for annotating library
    # classes such as subprocess.Popen.
    default_self = fill_typevars(info)
    if not is_new and not info.is_newtype:
        orig_self_types = [get_self_type(it, default_self) for it in signature.items]
    else:
        orig_self_types = [None] * len(signature.items)

    # The __init__ method might come from a generic superclass 'def_info'
    # with type variables that do not map identically to the type variables of
    # the class 'info' being constructed. For example:
    #
    #   class A(Generic[T]):
    #       def __init__(self, x: T) -> None: ...
    #   class B(A[List[T]]):
    #      ...
    #
    # We need to map B's __init__ to the type (List[T]) -> None.
    signature = bind_self(
        signature,
        original_type=default_self,
        is_classmethod=is_new,
        # Explicit instance self annotations have special handling in class_callable(),
        # we don't need to bind any type variables in them if they are generic.
        ignore_instances=True,
    )
    signature = cast(FunctionLike, map_type_from_supertype(signature, info, def_info))

    special_sig: str | None = None
    if def_info.fullname == "builtins.dict":
        # Special signature!
        special_sig = "dict"

    if isinstance(signature, CallableType):
        return class_callable(signature, info, fallback, special_sig, is_new, orig_self_types[0])
    else:
        # Overloaded __init__/__new__.
        assert isinstance(signature, Overloaded)
        items: list[CallableType] = []
        for item, orig_self in zip(signature.items, orig_self_types):
            items.append(class_callable(item, info, fallback, special_sig, is_new, orig_self))
        return Overloaded(items)


def class_callable(
    init_type: CallableType,
    info: TypeInfo,
    type_type: Instance,
    special_sig: str | None,
    is_new: bool,
    orig_self_type: Type | None = None,
) -> CallableType:
    """Create a type object type based on the signature of __init__."""
    variables: list[TypeVarLikeType] = []
    variables.extend(info.defn.type_vars)
    variables.extend(init_type.variables)

    from mypy.subtypes import is_subtype

    init_ret_type = get_proper_type(init_type.ret_type)
    orig_self_type = get_proper_type(orig_self_type)
    default_ret_type = fill_typevars(info)
    explicit_type = init_ret_type if is_new else orig_self_type
    if (
        isinstance(explicit_type, (Instance, TupleType, UninhabitedType))
        # We have to skip protocols, because it can be a subtype of a return type
        # by accident. Like `Hashable` is a subtype of `object`. See #11799
        and isinstance(default_ret_type, Instance)
        and not default_ret_type.type.is_protocol
        # Only use the declared return type from __new__ or declared self in __init__
        # if it is actually returning a subtype of what we would return otherwise.
        and is_subtype(explicit_type, default_ret_type, ignore_type_params=True)
    ):
        ret_type: Type = explicit_type
    else:
        ret_type = default_ret_type

    callable_type = init_type.copy_modified(
        ret_type=ret_type,
        fallback=type_type,
        name=None,
        variables=variables,
        special_sig=special_sig,
    )
    c = callable_type.with_name(info.name)
    return c


def map_type_from_supertype(typ: Type, sub_info: TypeInfo, super_info: TypeInfo) -> Type:
    """Map type variables in a type defined in a supertype context to be valid
    in the subtype context. Assume that the result is unique; if more than
    one type is possible, return one of the alternatives.

    For example, assume

      class D(Generic[S]): ...
      class C(D[E[T]], Generic[T]): ...

    Now S in the context of D would be mapped to E[T] in the context of C.
    """
    # Create the type of self in subtype, of form t[a1, ...].
    inst_type = fill_typevars(sub_info)
    if isinstance(inst_type, TupleType):
        inst_type = tuple_fallback(inst_type)
    # Map the type of self to supertype. This gets us a description of the
    # supertype type variables in terms of subtype variables, i.e. t[t1, ...]
    # so that any type variables in tN are to be interpreted in subtype
    # context.
    inst_type = map_instance_to_supertype(inst_type, super_info)
    # Finally expand the type variables in type with those in the previously
    # constructed type. Note that both type and inst_type may have type
    # variables, but in type they are interpreted in supertype context while
    # in inst_type they are interpreted in subtype context. This works even if
    # the names of type variables in supertype and subtype overlap.
    return expand_type_by_instance(typ, inst_type)


def supported_self_type(
    typ: ProperType, allow_callable: bool = True, allow_instances: bool = True
) -> bool:
    """Is this a supported kind of explicit self-types?

    Currently, this means an X or Type[X], where X is an instance or
    a type variable with an instance upper bound.
    """
    if isinstance(typ, TypeType):
        return supported_self_type(typ.item)
    if allow_callable and isinstance(typ, CallableType):
        # Special case: allow class callable instead of Type[...] as cls annotation,
        # as well as callable self for callback protocols.
        return True
    return isinstance(typ, TypeVarType) or (
        allow_instances and isinstance(typ, Instance) and typ != fill_typevars(typ.type)
    )


F = TypeVar("F", bound=FunctionLike)


def bind_self(
    method: F,
    original_type: Type | None = None,
    is_classmethod: bool = False,
    ignore_instances: bool = False,
) -> F:
    """Return a copy of `method`, with the type of its first parameter (usually
    self or cls) bound to original_type.

    If the type of `self` is a generic type (T, or Type[T] for classmethods),
    instantiate every occurrence of type with original_type in the rest of the
    signature and in the return type.

    original_type is the type of E in the expression E.copy(). It is None in
    compatibility checks. In this case we treat it as the erasure of the
    declared type of self.

    This way we can express "the type of self". For example:

    T = TypeVar('T', bound='A')
    class A:
        def copy(self: T) -> T: ...

    class B(A): pass

    b = B().copy()  # type: B

    """
    if isinstance(method, Overloaded):
        items = []
        original_type = get_proper_type(original_type)
        for c in method.items:
            if isinstance(original_type, Instance):
                # Filter based on whether declared self type can match actual object type.
                # For example, if self has type C[int] and method is accessed on a C[str] value,
                # omit this item. This is best effort since bind_self can be called in many
                # contexts, and doing complete validation might trigger infinite recursion.
                #
                # Note that overload item filtering normally happens elsewhere. This is needed
                # at least during constraint inference.
                keep = is_valid_self_type_best_effort(c, original_type)
            else:
                keep = True
            if keep:
                items.append(bind_self(c, original_type, is_classmethod, ignore_instances))
        if len(items) == 0:
            # If no item matches, returning all items helps avoid some spurious errors
            items = [
                bind_self(c, original_type, is_classmethod, ignore_instances) for c in method.items
            ]
        return cast(F, Overloaded(items))
    assert isinstance(method, CallableType)
    func: CallableType = method
    if not func.arg_types:
        # Invalid method, return something.
        return method
    if func.arg_kinds[0] in (ARG_STAR, ARG_STAR2):
        # The signature is of the form 'def foo(*args, ...)'.
        # In this case we shouldn't drop the first arg,
        # since func will be absorbed by the *args.
        # TODO: infer bounds on the type of *args?

        # In the case of **kwargs we should probably emit an error, but
        # for now we simply skip it, to avoid crashes down the line.
        return method
    self_param_type = get_proper_type(func.arg_types[0])

    variables: Sequence[TypeVarLikeType]
    # Having a def __call__(self: Callable[...], ...) can cause infinite recursion. Although
    # this special-casing looks not very principled, there is nothing meaningful we can infer
    # from such definition, since it is inherently indefinitely recursive.
    allow_callable = func.name is None or not func.name.startswith("__call__ of")
    if func.variables and supported_self_type(
        self_param_type, allow_callable=allow_callable, allow_instances=not ignore_instances
    ):
        from mypy.infer import infer_type_arguments

        if original_type is None:
            # TODO: type check method override (see #7861).
            original_type = erase_to_bound(self_param_type)
        original_type = get_proper_type(original_type)

        # Find which of method type variables appear in the type of "self".
        self_ids = {tv.id for tv in get_all_type_vars(self_param_type)}
        self_vars = [tv for tv in func.variables if tv.id in self_ids]

        # Solve for these type arguments using the actual class or instance type.
        typeargs = infer_type_arguments(
            self_vars, self_param_type, original_type, is_supertype=True
        )
        if (
            is_classmethod
            and any(isinstance(get_proper_type(t), UninhabitedType) for t in typeargs)
            and isinstance(original_type, (Instance, TypeVarType, TupleType))
        ):
            # In case we call a classmethod through an instance x, fallback to type(x).
            typeargs = infer_type_arguments(
                self_vars, self_param_type, TypeType(original_type), is_supertype=True
            )

        # Update the method signature with the solutions found.
        # Technically, some constraints might be unsolvable, make them Never.
        to_apply = [t if t is not None else UninhabitedType() for t in typeargs]
        func = expand_type(func, {tv.id: arg for tv, arg in zip(self_vars, to_apply)})
        variables = [v for v in func.variables if v not in self_vars]
    else:
        variables = func.variables

    original_type = get_proper_type(original_type)
    if isinstance(original_type, CallableType) and original_type.is_type_obj():
        original_type = TypeType.make_normalized(original_type.ret_type)
    res = func.copy_modified(
        arg_types=func.arg_types[1:],
        arg_kinds=func.arg_kinds[1:],
        arg_names=func.arg_names[1:],
        variables=variables,
        is_bound=True,
    )
    return cast(F, res)


def is_valid_self_type_best_effort(c: CallableType, self_type: Instance) -> bool:
    """Quickly check if self_type might match the self in a callable.

    Avoid performing any complex type operations. This is performance-critical.

    Default to returning True if we don't know (or it would be too expensive).
    """
    if (
        self_type.args
        and c.arg_types
        and isinstance((arg_type := get_proper_type(c.arg_types[0])), Instance)
        and c.arg_kinds[0] in (ARG_POS, ARG_OPT)
        and arg_type.args
        and self_type.type.fullname != "functools._SingleDispatchCallable"
    ):
        if self_type.type is not arg_type.type:
            # We can't map to supertype, since it could trigger expensive checks for
            # protocol types, so we consevatively assume this is fine.
            return True

        # Fast path: no explicit annotation on self
        if all(
            (
                type(arg) is TypeVarType
                and type(arg.upper_bound) is Instance
                and arg.upper_bound.type.fullname == "builtins.object"
            )
            for arg in arg_type.args
        ):
            return True

        from mypy.meet import is_overlapping_types

        return is_overlapping_types(self_type, c.arg_types[0])
    return True


def erase_to_bound(t: Type) -> Type:
    # TODO: use value restrictions to produce a union?
    t = get_proper_type(t)
    if isinstance(t, TypeVarType):
        return t.upper_bound
    if isinstance(t, TypeType):
        if isinstance(t.item, TypeVarType):
            return TypeType.make_normalized(t.item.upper_bound)
    return t


def callable_corresponding_argument(
    typ: NormalizedCallableType | Parameters, model: FormalArgument
) -> FormalArgument | None:
    """Return the argument a function that corresponds to `model`"""

    by_name = typ.argument_by_name(model.name)
    by_pos = typ.argument_by_position(model.pos)
    if by_name is None and by_pos is None:
        return None
    if by_name is not None and by_pos is not None:
        if by_name == by_pos:
            return by_name
        # If we're dealing with an optional pos-only and an optional
        # name-only arg, merge them.  This is the case for all functions
        # taking both *args and **args, or a pair of functions like so:

        # def right(a: int = ...) -> None: ...
        # def left(__a: int = ..., *, a: int = ...) -> None: ...
        from mypy.subtypes import is_equivalent

        if (
            not (by_name.required or by_pos.required)
            and by_pos.name is None
            and by_name.pos is None
            and is_equivalent(by_name.typ, by_pos.typ)
        ):
            return FormalArgument(by_name.name, by_pos.pos, by_name.typ, False)
    return by_name if by_name is not None else by_pos


def simple_literal_type(t: ProperType | None) -> Instance | None:
    """Extract the underlying fallback Instance type for a simple Literal"""
    if isinstance(t, Instance) and t.last_known_value is not None:
        t = t.last_known_value
    if isinstance(t, LiteralType):
        return t.fallback
    return None


def is_simple_literal(t: ProperType) -> bool:
    if isinstance(t, LiteralType):
        return t.fallback.type.is_enum or t.fallback.type.fullname == "builtins.str"
    if isinstance(t, Instance):
        return t.last_known_value is not None and isinstance(t.last_known_value.value, str)
    return False


def make_simplified_union(
    items: Sequence[Type],
    line: int = -1,
    column: int = -1,
    *,
    keep_erased: bool = False,
    contract_literals: bool = True,
    handle_recursive: bool = True,
) -> ProperType:
    """Build union type with redundant union items removed.

    If only a single item remains, this may return a non-union type.

    Examples:

    * [int, str] -> Union[int, str]
    * [int, object] -> object
    * [int, int] -> int
    * [int, Any] -> Union[int, Any] (Any types are not simplified away!)
    * [Any, Any] -> Any
    * [int, Union[bytes, str]] -> Union[int, bytes, str]

    Note: This must NOT be used during semantic analysis, since TypeInfos may not
          be fully initialized.

    The keep_erased flag is used for type inference against union types
    containing type variables. If set to True, keep all ErasedType items.

    The contract_literals flag indicates whether we need to contract literal types
    back into a sum type. Set it to False when called by try_expanding_sum_type_
    to_union().
    """
    # Step 1: expand all nested unions
    items = flatten_nested_unions(items, handle_recursive=handle_recursive)

    # Step 2: fast path for single item
    if len(items) == 1:
        return get_proper_type(items[0])

    # Step 3: remove redundant unions
    simplified_set: Sequence[Type] = _remove_redundant_union_items(items, keep_erased)

    # Step 4: If more than one literal exists in the union, try to simplify
    if (
        contract_literals
        and sum(isinstance(get_proper_type(item), LiteralType) for item in simplified_set) > 1
    ):
        simplified_set = try_contracting_literals_in_union(simplified_set)

    result = get_proper_type(UnionType.make_union(simplified_set, line, column))

    nitems = len(items)
    if nitems > 1 and (
        nitems > 2 or not (type(items[0]) is NoneType or type(items[1]) is NoneType)
    ):
        # Step 5: At last, we erase any (inconsistent) extra attributes on instances.

        # Initialize with None instead of an empty set as a micro-optimization. The set
        # is needed very rarely, so we try to avoid constructing it.
        extra_attrs_set: set[ExtraAttrs] | None = None
        for item in items:
            instance = try_getting_instance_fallback(item)
            if instance and instance.extra_attrs:
                if extra_attrs_set is None:
                    extra_attrs_set = {instance.extra_attrs}
                else:
                    extra_attrs_set.add(instance.extra_attrs)

        if extra_attrs_set is not None and len(extra_attrs_set) > 1:
            fallback = try_getting_instance_fallback(result)
            if fallback:
                fallback.extra_attrs = None

    return result


def _remove_redundant_union_items(items: list[Type], keep_erased: bool) -> list[Type]:
    from mypy.subtypes import is_proper_subtype

    # The first pass through this loop, we check if later items are subtypes of earlier items.
    # The second pass through this loop, we check if earlier items are subtypes of later items
    # (by reversing the remaining items)
    for _direction in range(2):
        new_items: list[Type] = []
        # seen is a map from a type to its index in new_items
        seen: dict[ProperType, int] = {}
        unduplicated_literal_fallbacks: set[Instance] | None = None
        for ti in items:
            proper_ti = get_proper_type(ti)

            # UninhabitedType is always redundant
            if isinstance(proper_ti, UninhabitedType):
                continue

            duplicate_index = -1
            # Quickly check if we've seen this type
            if proper_ti in seen:
                duplicate_index = seen[proper_ti]
            elif (
                isinstance(proper_ti, LiteralType)
                and unduplicated_literal_fallbacks is not None
                and proper_ti.fallback in unduplicated_literal_fallbacks
            ):
                # This is an optimisation for unions with many LiteralType
                # We've already checked for exact duplicates. This means that any super type of
                # the LiteralType must be a super type of its fallback. If we've gone through
                # the expensive loop below and found no super type for a previous LiteralType
                # with the same fallback, we can skip doing that work again and just add the type
                # to new_items
                pass
            else:
                # If not, check if we've seen a supertype of this type
                for j, tj in enumerate(new_items):
                    tj = get_proper_type(tj)
                    # If tj is an Instance with a last_known_value, do not remove proper_ti
                    # (unless it's an instance with the same last_known_value)
                    if (
                        isinstance(tj, Instance)
                        and tj.last_known_value is not None
                        and not (
                            isinstance(proper_ti, Instance)
                            and tj.last_known_value == proper_ti.last_known_value
                        )
                    ):
                        continue

                    if is_proper_subtype(
                        ti, tj, keep_erased_types=keep_erased, ignore_promotions=True
                    ):
                        duplicate_index = j
                        break
            if duplicate_index != -1:
                # If deleted subtypes had more general truthiness, use that
                orig_item = new_items[duplicate_index]
                if not orig_item.can_be_true and ti.can_be_true:
                    new_items[duplicate_index] = true_or_false(orig_item)
                elif not orig_item.can_be_false and ti.can_be_false:
                    new_items[duplicate_index] = true_or_false(orig_item)
            else:
                # We have a non-duplicate item, add it to new_items
                seen[proper_ti] = len(new_items)
                new_items.append(ti)
                if isinstance(proper_ti, LiteralType):
                    if unduplicated_literal_fallbacks is None:
                        unduplicated_literal_fallbacks = set()
                    unduplicated_literal_fallbacks.add(proper_ti.fallback)

        items = new_items
        if len(items) <= 1:
            break
        items.reverse()

    return items


def _get_type_method_ret_type(t: ProperType, *, name: str) -> Type | None:
    # For Enum literals the ret_type can change based on the Enum
    # we need to check the type of the enum rather than the literal
    if isinstance(t, LiteralType) and t.is_enum_literal():
        t = t.fallback

    if isinstance(t, Instance):
        sym = t.type.get(name)
        if sym:
            sym_type = get_proper_type(sym.type)
            if isinstance(sym_type, CallableType):
                return sym_type.ret_type

    return None


def true_only(t: Type) -> ProperType:
    """
    Restricted version of t with only True-ish values
    """
    t = get_proper_type(t)

    if not t.can_be_true:
        # All values of t are False-ish, so there are no true values in it
        return UninhabitedType(line=t.line, column=t.column)
    elif not t.can_be_false:
        # All values of t are already True-ish, so true_only is idempotent in this case
        return t
    elif isinstance(t, UnionType):
        # The true version of a union type is the union of the true versions of its components
        new_items = [true_only(item) for item in t.items]
        can_be_true_items = [item for item in new_items if item.can_be_true]
        return make_simplified_union(can_be_true_items, line=t.line, column=t.column)
    else:
        ret_type = _get_type_method_ret_type(t, name="__bool__") or _get_type_method_ret_type(
            t, name="__len__"
        )

        if ret_type and not ret_type.can_be_true:
            return UninhabitedType(line=t.line, column=t.column)

        new_t = copy_type(t)
        new_t.can_be_false = False
        return new_t


def false_only(t: Type) -> ProperType:
    """
    Restricted version of t with only False-ish values
    """
    t = get_proper_type(t)

    if not t.can_be_false:
        if state.strict_optional:
            # All values of t are True-ish, so there are no false values in it
            return UninhabitedType(line=t.line)
        else:
            # When strict optional checking is disabled, everything can be
            # False-ish since anything can be None
            return NoneType(line=t.line)
    elif not t.can_be_true:
        # All values of t are already False-ish, so false_only is idempotent in this case
        return t
    elif isinstance(t, UnionType):
        # The false version of a union type is the union of the false versions of its components
        new_items = [false_only(item) for item in t.items]
        can_be_false_items = [item for item in new_items if item.can_be_false]
        return make_simplified_union(can_be_false_items, line=t.line, column=t.column)
    elif isinstance(t, Instance) and t.type.fullname in ("builtins.str", "builtins.bytes"):
        return LiteralType("", fallback=t)
    elif isinstance(t, Instance) and t.type.fullname == "builtins.int":
        return LiteralType(0, fallback=t)
    else:
        ret_type = _get_type_method_ret_type(t, name="__bool__") or _get_type_method_ret_type(
            t, name="__len__"
        )

        if ret_type:
            if not ret_type.can_be_false:
                return UninhabitedType(line=t.line)
        elif isinstance(t, Instance):
            if t.type.is_final or t.type.is_enum:
                return UninhabitedType(line=t.line)
        elif isinstance(t, LiteralType) and t.is_enum_literal():
            return UninhabitedType(line=t.line)

        new_t = copy_type(t)
        new_t.can_be_true = False
        return new_t


def true_or_false(t: Type) -> ProperType:
    """
    Unrestricted version of t with both True-ish and False-ish values
    """
    t = get_proper_type(t)

    if isinstance(t, UnionType):
        new_items = [true_or_false(item) for item in t.items]
        return make_simplified_union(new_items, line=t.line, column=t.column)

    new_t = copy_type(t)
    new_t.can_be_true = new_t.can_be_true_default()
    new_t.can_be_false = new_t.can_be_false_default()
    return new_t


def erase_def_to_union_or_bound(tdef: TypeVarLikeType) -> Type:
    # TODO(PEP612): fix for ParamSpecType
    if isinstance(tdef, ParamSpecType):
        return AnyType(TypeOfAny.from_error)
    if isinstance(tdef, TypeVarType) and tdef.values:
        return make_simplified_union(tdef.values)
    else:
        return tdef.upper_bound


def erase_to_union_or_bound(typ: TypeVarType) -> ProperType:
    if typ.values:
        return make_simplified_union(typ.values)
    else:
        return get_proper_type(typ.upper_bound)


def function_type(func: FuncBase, fallback: Instance) -> FunctionLike:
    if func.type:
        assert isinstance(func.type, FunctionLike)
        return func.type
    else:
        # Implicit type signature with dynamic types.
        if isinstance(func, FuncItem):
            return callable_type(func, fallback)
        else:
            # Broken overloads can have self.type set to None.
            # TODO: should we instead always set the type in semantic analyzer?
            assert isinstance(func, OverloadedFuncDef)
            any_type = AnyType(TypeOfAny.from_error)
            dummy = CallableType(
                [any_type, any_type],
                [ARG_STAR, ARG_STAR2],
                [None, None],
                any_type,
                fallback,
                line=func.line,
                is_ellipsis_args=True,
            )
            # Return an Overloaded, because some callers may expect that
            # an OverloadedFuncDef has an Overloaded type.
            return Overloaded([dummy])


def callable_type(
    fdef: FuncItem, fallback: Instance, ret_type: Type | None = None
) -> CallableType:
    # TODO: somewhat unfortunate duplication with prepare_method_signature in semanal
    if fdef.info and (not fdef.is_static or fdef.name == "__new__") and fdef.arg_names:
        self_type: Type = fill_typevars(fdef.info)
        if fdef.is_class or fdef.name == "__new__":
            self_type = TypeType.make_normalized(self_type)
        args = [self_type] + [AnyType(TypeOfAny.unannotated)] * (len(fdef.arg_names) - 1)
    else:
        args = [AnyType(TypeOfAny.unannotated)] * len(fdef.arg_names)

    return CallableType(
        args,
        fdef.arg_kinds,
        fdef.arg_names,
        ret_type or AnyType(TypeOfAny.unannotated),
        fallback,
        name=fdef.name,
        line=fdef.line,
        column=fdef.column,
        implicit=True,
        # We need this for better error messages, like missing `self` note:
        definition=fdef if isinstance(fdef, FuncDef) else None,
    )


def try_getting_str_literals(expr: Expression, typ: Type) -> list[str] | None:
    """If the given expression or type corresponds to a string literal
    or a union of string literals, returns a list of the underlying strings.
    Otherwise, returns None.

    Specifically, this function is guaranteed to return a list with
    one or more strings if one of the following is true:

    1. 'expr' is a StrExpr
    2. 'typ' is a LiteralType containing a string
    3. 'typ' is a UnionType containing only LiteralType of strings
    """
    if isinstance(expr, StrExpr):
        return [expr.value]

    # TODO: See if we can eliminate this function and call the below one directly
    return try_getting_str_literals_from_type(typ)


def try_getting_str_literals_from_type(typ: Type) -> list[str] | None:
    """If the given expression or type corresponds to a string Literal
    or a union of string Literals, returns a list of the underlying strings.
    Otherwise, returns None.

    For example, if we had the type 'Literal["foo", "bar"]' as input, this function
    would return a list of strings ["foo", "bar"].
    """
    return try_getting_literals_from_type(typ, str, "builtins.str")


def try_getting_int_literals_from_type(typ: Type) -> list[int] | None:
    """If the given expression or type corresponds to an int Literal
    or a union of int Literals, returns a list of the underlying ints.
    Otherwise, returns None.

    For example, if we had the type 'Literal[1, 2, 3]' as input, this function
    would return a list of ints [1, 2, 3].
    """
    return try_getting_literals_from_type(typ, int, "builtins.int")


T = TypeVar("T")


def try_getting_literals_from_type(
    typ: Type, target_literal_type: type[T], target_fullname: str
) -> list[T] | None:
    """If the given expression or type corresponds to a Literal or
    union of Literals where the underlying values correspond to the given
    target type, returns a list of those underlying values. Otherwise,
    returns None.
    """
    typ = get_proper_type(typ)

    if isinstance(typ, Instance) and typ.last_known_value is not None:
        possible_literals: list[Type] = [typ.last_known_value]
    elif isinstance(typ, UnionType):
        possible_literals = list(typ.items)
    else:
        possible_literals = [typ]

    literals: list[T] = []
    for lit in get_proper_types(possible_literals):
        if isinstance(lit, LiteralType) and lit.fallback.type.fullname == target_fullname:
            val = lit.value
            if isinstance(val, target_literal_type):
                literals.append(val)
            else:
                return None
        else:
            return None
    return literals


def is_literal_type_like(t: Type | None) -> bool:
    """Returns 'true' if the given type context is potentially either a LiteralType,
    a Union of LiteralType, or something similar.
    """
    t = get_proper_type(t)
    if t is None:
        return False
    elif isinstance(t, LiteralType):
        return True
    elif isinstance(t, UnionType):
        return any(is_literal_type_like(item) for item in t.items)
    elif isinstance(t, TypeVarType):
        return is_literal_type_like(t.upper_bound) or any(
            is_literal_type_like(item) for item in t.values
        )
    else:
        return False


def is_singleton_type(typ: Type) -> bool:
    """Returns 'true' if this type is a "singleton type" -- if there exists
    exactly only one runtime value associated with this type.

    That is, given two values 'a' and 'b' that have the same type 't',
    'is_singleton_type(t)' returns True if and only if the expression 'a is b' is
    always true.

    Currently, this returns True when given NoneTypes, enum LiteralTypes,
    enum types with a single value and ... (Ellipses).

    Note that other kinds of LiteralTypes cannot count as singleton types. For
    example, suppose we do 'a = 100000 + 1' and 'b = 100001'. It is not guaranteed
    that 'a is b' will always be true -- some implementations of Python will end up
    constructing two distinct instances of 100001.
    """
    typ = get_proper_type(typ)
    return typ.is_singleton_type()


def try_expanding_sum_type_to_union(typ: Type, target_fullname: str) -> ProperType:
    """Attempts to recursively expand any enum Instances with the given target_fullname
    into a Union of all of its component LiteralTypes.

    For example, if we have:

        class Color(Enum):
            RED = 1
            BLUE = 2
            YELLOW = 3

        class Status(Enum):
            SUCCESS = 1
            FAILURE = 2
            UNKNOWN = 3

    ...and if we call `try_expanding_sum_type_to_union(Union[Color, Status], 'module.Color')`,
    this function will return Literal[Color.RED, Color.BLUE, Color.YELLOW, Status].
    """
    typ = get_proper_type(typ)

    if isinstance(typ, UnionType):
        items = [
            try_expanding_sum_type_to_union(item, target_fullname) for item in typ.relevant_items()
        ]
        return make_simplified_union(items, contract_literals=False)

    if isinstance(typ, Instance) and typ.type.fullname == target_fullname:
        if typ.type.fullname == "builtins.bool":
            items = [LiteralType(True, typ), LiteralType(False, typ)]
            return make_simplified_union(items, contract_literals=False)

        if typ.type.is_enum:
            items = [LiteralType(name, typ) for name in typ.type.enum_members]
            if not items:
                return typ
            return make_simplified_union(items, contract_literals=False)

    return typ


def try_contracting_literals_in_union(types: Sequence[Type]) -> list[ProperType]:
    """Contracts any literal types back into a sum type if possible.

    Requires a flattened union and does not descend into children.

    Will replace the first instance of the literal with the sum type and
    remove all others.

    If we call `try_contracting_union(Literal[Color.RED, Color.BLUE, Color.YELLOW])`,
    this function will return Color.

    We also treat `Literal[True, False]` as `bool`.
    """
    proper_types = [get_proper_type(typ) for typ in types]
    sum_types: dict[str, tuple[set[Any], list[int]]] = {}
    marked_for_deletion = set()
    for idx, typ in enumerate(proper_types):
        if isinstance(typ, LiteralType):
            fullname = typ.fallback.type.fullname
            if typ.fallback.type.is_enum or isinstance(typ.value, bool):
                if fullname not in sum_types:
                    sum_types[fullname] = (
                        (
                            set(typ.fallback.type.enum_members)
                            if typ.fallback.type.is_enum
                            else {True, False}
                        ),
                        [],
                    )
                literals, indexes = sum_types[fullname]
                literals.discard(typ.value)
                indexes.append(idx)
                if not literals:
                    first, *rest = indexes
                    proper_types[first] = typ.fallback
                    marked_for_deletion |= set(rest)
    return list(
        itertools.compress(
            proper_types, [(i not in marked_for_deletion) for i in range(len(proper_types))]
        )
    )


def coerce_to_literal(typ: Type) -> Type:
    """Recursively converts any Instances that have a last_known_value or are
    instances of enum types with a single value into the corresponding LiteralType.
    """
    original_type = typ
    typ = get_proper_type(typ)
    if isinstance(typ, UnionType):
        new_items = [coerce_to_literal(item) for item in typ.items]
        return UnionType.make_union(new_items)
    elif isinstance(typ, Instance):
        if typ.last_known_value:
            return typ.last_known_value
        elif typ.type.is_enum:
            enum_values = typ.type.enum_members
            if len(enum_values) == 1:
                return LiteralType(value=enum_values[0], fallback=typ)
    return original_type


def get_type_vars(tp: Type) -> list[TypeVarType]:
    return cast("list[TypeVarType]", tp.accept(TypeVarExtractor()))


def get_all_type_vars(tp: Type) -> list[TypeVarLikeType]:
    # TODO: should we always use this function instead of get_type_vars() above?
    return tp.accept(TypeVarExtractor(include_all=True))


class TypeVarExtractor(TypeQuery[list[TypeVarLikeType]]):
    def __init__(self, include_all: bool = False) -> None:
        super().__init__(self._merge)
        self.include_all = include_all

    def _merge(self, iter: Iterable[list[TypeVarLikeType]]) -> list[TypeVarLikeType]:
        out = []
        for item in iter:
            out.extend(item)
        return out

    def visit_type_var(self, t: TypeVarType) -> list[TypeVarLikeType]:
        return [t]

    def visit_param_spec(self, t: ParamSpecType) -> list[TypeVarLikeType]:
        return [t] if self.include_all else []

    def visit_type_var_tuple(self, t: TypeVarTupleType) -> list[TypeVarLikeType]:
        return [t] if self.include_all else []


def freeze_all_type_vars(member_type: Type) -> None:
    member_type.accept(FreezeTypeVarsVisitor())


class FreezeTypeVarsVisitor(TypeTraverserVisitor):
    def visit_callable_type(self, t: CallableType) -> None:
        for v in t.variables:
            v.id.meta_level = 0
        super().visit_callable_type(t)


def custom_special_method(typ: Type, name: str, check_all: bool = False) -> bool:
    """Does this type have a custom special method such as __format__() or __eq__()?

    If check_all is True ensure all items of a union have a custom method, not just some.
    """
    typ = get_proper_type(typ)
    if isinstance(typ, Instance):
        method = typ.type.get(name)
        if method and isinstance(method.node, (SYMBOL_FUNCBASE_TYPES, Decorator, Var)):
            if method.node.info:
                return not method.node.info.fullname.startswith(("builtins.", "typing."))
        return False
    if isinstance(typ, UnionType):
        if check_all:
            return all(custom_special_method(t, name, check_all) for t in typ.items)
        return any(custom_special_method(t, name) for t in typ.items)
    if isinstance(typ, TupleType):
        return custom_special_method(tuple_fallback(typ), name, check_all)
    if isinstance(typ, FunctionLike) and typ.is_type_obj():
        # Look up __method__ on the metaclass for class objects.
        return custom_special_method(typ.fallback, name, check_all)
    if isinstance(typ, TypeType) and isinstance(typ.item, Instance):
        if typ.item.type.metaclass_type:
            # Look up __method__ on the metaclass for class objects.
            return custom_special_method(typ.item.type.metaclass_type, name, check_all)
    if isinstance(typ, AnyType):
        # Avoid false positives in uncertain cases.
        return True
    # TODO: support other types (see ExpressionChecker.has_member())?
    return False


def separate_union_literals(t: UnionType) -> tuple[Sequence[LiteralType], Sequence[Type]]:
    """Separate literals from other members in a union type."""
    literal_items = []
    union_items = []

    for item in t.items:
        proper = get_proper_type(item)
        if isinstance(proper, LiteralType):
            literal_items.append(proper)
        else:
            union_items.append(item)

    return literal_items, union_items


def try_getting_instance_fallback(typ: Type) -> Instance | None:
    """Returns the Instance fallback for this type if one exists or None."""
    typ = get_proper_type(typ)
    if isinstance(typ, Instance):
        return typ
    elif isinstance(typ, LiteralType):
        return typ.fallback
    elif isinstance(typ, NoneType):
        return None  # Fast path for None, which is common
    elif isinstance(typ, FunctionLike):
        return typ.fallback
    elif isinstance(typ, TupleType):
        return typ.partial_fallback
    elif isinstance(typ, TypedDictType):
        return typ.fallback
    elif isinstance(typ, TypeVarType):
        return try_getting_instance_fallback(typ.upper_bound)
    return None


def fixup_partial_type(typ: Type) -> Type:
    """Convert a partial type that we couldn't resolve into something concrete.

    This means, for None we make it Optional[Any], and for anything else we
    fill in all of the type arguments with Any.
    """
    if not isinstance(typ, PartialType):
        return typ
    if typ.type is None:
        return UnionType.make_union([AnyType(TypeOfAny.unannotated), NoneType()])
    else:
        return Instance(typ.type, [AnyType(TypeOfAny.unannotated)] * len(typ.type.type_vars))


def get_protocol_member(
    left: Instance, member: str, class_obj: bool, is_lvalue: bool = False
) -> Type | None:
    if member == "__call__" and class_obj:
        # Special case: class objects always have __call__ that is just the constructor.

        def named_type(fullname: str) -> Instance:
            return Instance(left.type.mro[-1], [])

        return type_object_type(left.type, named_type)

    if member == "__call__" and left.type.is_metaclass(precise=True):
        # Special case: we want to avoid falling back to metaclass __call__
        # if constructor signature didn't match, this can cause many false negatives.
        return None

    from mypy.subtypes import find_member

    subtype = find_member(member, left, left, class_obj=class_obj, is_lvalue=is_lvalue)
    if isinstance(subtype, PartialType):
        subtype = (
            NoneType()
            if subtype.type is None
            else Instance(
                subtype.type, [AnyType(TypeOfAny.unannotated)] * len(subtype.type.type_vars)
            )
        )
    return subtype



"""
Cloud Browser - PyCloud OS Web Tarayıcısı
WebEngine tabanlı modern sekmeli tarayıcı, PDF desteği ve yer imi yönetimi
"""

import sys
import os
import json
import logging
import argparse
from typing import Dict, List, Optional, Any
from datetime import datetime
from pathlib import Path
from urllib.parse import urlparse, urljoin

# PyQt6 import with fallback
try:
    from PyQt6.QtWidgets import *
    from PyQt6.QtCore import *
    from PyQt6.QtGui import *
    from PyQt6.QtWebEngineWidgets import *
    from PyQt6.QtWebEngineCore import *
    PYQT_AVAILABLE = True
    WEBENGINE_AVAILABLE = True
except ImportError:
    try:
        from PyQt6.QtWidgets import *
        from PyQt6.QtCore import *
        from PyQt6.QtGui import *
        PYQT_AVAILABLE = True
        WEBENGINE_AVAILABLE = False
        print("PyQt6 WebEngine bulunamadı - Basit tarayıcı modu")
    except ImportError:
        PYQT_AVAILABLE = False
        WEBENGINE_AVAILABLE = False
        print("PyQt6 bulunamadı - Browser text modunda çalışacak")

class BrowserTab:
    """Tarayıcı sekmesi"""
    
    def __init__(self, title: str = "Yeni Sekme", url: str = "about:blank"):
        self.title = title
        self.url = url
        self.history: List[str] = []
        self.history_index = -1
        self.is_loading = False
        self.favicon = None
        self.created_at = datetime.now()
        
    def add_to_history(self, url: str):
        """Geçmişe URL ekle"""
        if url and url != "about:blank":
            # Mevcut pozisyondan sonrasını sil
            self.history = self.history[:self.history_index + 1]
            # Yeni URL'yi ekle
            self.history.append(url)
            self.history_index = len(self.history) - 1
    
    def can_go_back(self) -> bool:
        """Geri gidebilir mi?"""
        return self.history_index > 0
    
    def can_go_forward(self) -> bool:
        """İleri gidebilir mi?"""
        return self.history_index < len(self.history) - 1
    
    def go_back(self) -> Optional[str]:
        """Geri git"""
        if self.can_go_back():
            self.history_index -= 1
            return self.history[self.history_index]
        return None
    
    def go_forward(self) -> Optional[str]:
        """İleri git"""
        if self.can_go_forward():
            self.history_index += 1
            return self.history[self.history_index]
        return None

class BookmarkManager:
    """Yer imi yöneticisi"""
    
    def __init__(self):
        self.bookmarks: List[Dict[str, Any]] = []
        self.bookmarks_file = "system/browser_bookmarks.json"
        self.load_bookmarks()
    
    def load_bookmarks(self):
        """Yer imlerini yükle"""
        try:
            if os.path.exists(self.bookmarks_file):
                with open(self.bookmarks_file, 'r', encoding='utf-8') as f:
                    self.bookmarks = json.load(f)
        except Exception as e:
            print(f"Yer imi yükleme hatası: {e}")
            self.bookmarks = []
    
    def save_bookmarks(self):
        """Yer imlerini kaydet"""
        try:
            os.makedirs(os.path.dirname(self.bookmarks_file), exist_ok=True)
            with open(self.bookmarks_file, 'w', encoding='utf-8') as f:
                json.dump(self.bookmarks, f, indent=2, ensure_ascii=False, default=str)
        except Exception as e:
            print(f"Yer imi kaydetme hatası: {e}")
    
    def add_bookmark(self, title: str, url: str, folder: str = "Genel"):
        """Yer imi ekle"""
        bookmark = {
            "title": title,
            "url": url,
            "folder": folder,
            "created": datetime.now().isoformat(),
            "visits": 0
        }
        self.bookmarks.append(bookmark)
        self.save_bookmarks()
    
    def remove_bookmark(self, url: str):
        """Yer imi sil"""
        self.bookmarks = [b for b in self.bookmarks if b["url"] != url]
        self.save_bookmarks()
    
    def get_bookmarks(self, folder: Optional[str] = None) -> List[Dict[str, Any]]:
        """Yer imlerini al"""
        if folder:
            return [b for b in self.bookmarks if b.get("folder") == folder]
        return self.bookmarks
    
    def is_bookmarked(self, url: str) -> bool:
        """URL yer imi var mı?"""
        return any(b["url"] == url for b in self.bookmarks)

if PYQT_AVAILABLE and WEBENGINE_AVAILABLE:
    class BrowserWindow(QMainWindow):
        """Ana tarayıcı penceresi"""
        
        def __init__(self, kernel=None):
            super().__init__()
            self.kernel = kernel
            self.tabs: Dict[int, BrowserTab] = {}
            self.current_tab_index = -1
            self.bookmark_manager = BookmarkManager()
            
            self.init_ui()
            self.apply_theme()
            self.new_tab("https://www.google.com")
        
        def init_ui(self):
            """UI'yı başlat"""
            self.setWindowTitle("PyCloud Browser")
            self.setGeometry(100, 100, 1200, 800)
            
            # Ana widget
            central_widget = QWidget()
            self.setCentralWidget(central_widget)
            
            layout = QVBoxLayout()
            
            # Navigasyon araç çubuğu
            nav_toolbar = self.create_navigation_toolbar()
            layout.addWidget(nav_toolbar)
            
            # Tab widget
            self.tab_widget = QTabWidget()
            self.tab_widget.setTabsClosable(True)
            self.tab_widget.tabCloseRequested.connect(self.close_tab)
            self.tab_widget.currentChanged.connect(self.tab_changed)
            layout.addWidget(self.tab_widget)
            
            # Durum çubuğu
            self.status_bar = QStatusBar()
            self.setStatusBar(self.status_bar)
            self.status_bar.showMessage("Hazır")
            
            central_widget.setLayout(layout)
            
            # Menü çubuğu
            self.create_menu_bar()
            
            # Kısayollar
            self.setup_shortcuts()
        
        def create_menu_bar(self):
            """Menü çubuğu oluştur"""
            menubar = self.menuBar()
            
            # Dosya menüsü
            file_menu = menubar.addMenu('Dosya')
            
            new_tab_action = QAction('Yeni Sekme', self)
            new_tab_action.setShortcut('Ctrl+T')
            new_tab_action.triggered.connect(lambda: self.new_tab())
            file_menu.addAction(new_tab_action)
            
            new_window_action = QAction('Yeni Pencere', self)
            new_window_action.setShortcut('Ctrl+N')
            new_window_action.triggered.connect(self.new_window)
            file_menu.addAction(new_window_action)
            
            file_menu.addSeparator()
            
            close_tab_action = QAction('Sekmeyi Kapat', self)
            close_tab_action.setShortcut('Ctrl+W')
            close_tab_action.triggered.connect(self.close_current_tab)
            file_menu.addAction(close_tab_action)
            
            exit_action = QAction('Çıkış', self)
            exit_action.setShortcut('Ctrl+Q')
            exit_action.triggered.connect(self.close)
            file_menu.addAction(exit_action)
            
            # Yer imleri menüsü
            bookmark_menu = menubar.addMenu('Yer İmleri')
            
            add_bookmark_action = QAction('Bu Sayfayı Yer İmine Ekle', self)
            add_bookmark_action.setShortcut('Ctrl+D')
            add_bookmark_action.triggered.connect(self.add_current_bookmark)
            bookmark_menu.addAction(add_bookmark_action)
            
            manage_bookmarks_action = QAction('Yer İmlerini Yönet', self)
            manage_bookmarks_action.triggered.connect(self.manage_bookmarks)
            bookmark_menu.addAction(manage_bookmarks_action)
            
            bookmark_menu.addSeparator()
            
            # Yer imi listesi
            self.update_bookmark_menu()
            
            # Görünüm menüsü
            view_menu = menubar.addMenu('Görünüm')
            
            reload_action = QAction('Yenile', self)
            reload_action.setShortcut('F5')
            reload_action.triggered.connect(self.reload_page)
            view_menu.addAction(reload_action)
            
            fullscreen_action = QAction('Tam Ekran', self)
            fullscreen_action.setShortcut('F11')
            fullscreen_action.triggered.connect(self.toggle_fullscreen)
            view_menu.addAction(fullscreen_action)
        
        def create_navigation_toolbar(self):
            """Navigasyon araç çubuğu oluştur"""
            toolbar = QToolBar()
            toolbar.setMovable(False)
            
            # Geri buton
            self.back_btn = QPushButton("◀")
            self.back_btn.setToolTip("Geri")
            self.back_btn.clicked.connect(self.go_back)
            self.back_btn.setEnabled(False)
            toolbar.addWidget(self.back_btn)
            
            # İleri buton
            self.forward_btn = QPushButton("▶")
            self.forward_btn.setToolTip("İleri")
            self.forward_btn.clicked.connect(self.go_forward)
            self.forward_btn.setEnabled(False)
            toolbar.addWidget(self.forward_btn)
            
            # Yenile buton
            self.reload_btn = QPushButton("🔄")
            self.reload_btn.setToolTip("Yenile")
            self.reload_btn.clicked.connect(self.reload_page)
            toolbar.addWidget(self.reload_btn)
            
            # Ana sayfa
            home_btn = QPushButton("🏠")
            home_btn.setToolTip("Ana Sayfa")
            home_btn.clicked.connect(self.go_home)
            toolbar.addWidget(home_btn)
            
            # Adres çubuğu
            self.address_bar = QLineEdit()
            self.address_bar.setPlaceholderText("Bir URL girin veya arama yapın...")
            self.address_bar.returnPressed.connect(self.navigate_to_url)
            toolbar.addWidget(self.address_bar)
            
            # Yer imi butonu
            self.bookmark_btn = QPushButton("⭐")
            self.bookmark_btn.setToolTip("Yer İmine Ekle")
            self.bookmark_btn.clicked.connect(self.toggle_bookmark)
            toolbar.addWidget(self.bookmark_btn)
            
            # Yeni sekme
            new_tab_btn = QPushButton("➕")
            new_tab_btn.setToolTip("Yeni Sekme")
            new_tab_btn.clicked.connect(lambda: self.new_tab())
            toolbar.addWidget(new_tab_btn)
            
            return toolbar
        
        def setup_shortcuts(self):
            """Kısayolları ayarla"""
            # Tab gezinme
            next_tab_shortcut = QShortcut(QKeySequence("Ctrl+Tab"), self)
            next_tab_shortcut.activated.connect(self.next_tab)
            
            prev_tab_shortcut = QShortcut(QKeySequence("Ctrl+Shift+Tab"), self)
            prev_tab_shortcut.activated.connect(self.prev_tab)
            
            # Yeniden aç
            reopen_shortcut = QShortcut(QKeySequence("Ctrl+Shift+T"), self)
            reopen_shortcut.activated.connect(self.reopen_closed_tab)
        
        def apply_theme(self):
            """Tema uygula"""
            self.setStyleSheet("""
                QMainWindow {
                    background-color: #f0f0f0;
                }
                QToolBar {
                    background-color: #ffffff;
                    border: 1px solid #d0d0d0;
                    spacing: 3px;
                    padding: 4px;
                }
                QPushButton {
                    background-color: #ffffff;
                    border: 1px solid #cccccc;
                    padding: 6px 12px;
                    border-radius: 4px;
                    font-size: 12px;
                }
                QPushButton:hover {
                    background-color: #e6e6e6;
                }
                QPushButton:pressed {
                    background-color: #d0d0d0;
                }
                QPushButton:disabled {
                    background-color: #f5f5f5;
                    color: #999999;
                }
                QLineEdit {
                    border: 1px solid #cccccc;
                    padding: 8px;
                    border-radius: 4px;
                    font-size: 12px;
                }
                QLineEdit:focus {
                    border: 2px solid #0078d4;
                }
                QTabWidget::pane {
                    border: 1px solid #cccccc;
                    background-color: #ffffff;
                }
                QTabBar::tab {
                    background-color: #f5f5f5;
                    padding: 8px 16px;
                    margin-right: 2px;
                    border-top-left-radius: 4px;
                    border-top-right-radius: 4px;
                    border: 1px solid #cccccc;
                    border-bottom: none;
                }
                QTabBar::tab:selected {
                    background-color: #ffffff;
                }
                QStatusBar {
                    background-color: #f5f5f5;
                    border-top: 1px solid #cccccc;
                }
            """)
        
        def new_tab(self, url: str = "about:blank"):
            """Yeni sekme oluştur"""
            # Tab bilgisi oluştur
            tab_info = BrowserTab("Yeni Sekme", url)
            
            # Web view oluştur
            web_view = QWebEngineView()
            
            # Sayfa yüklenme olayları
            web_view.titleChanged.connect(lambda title: self.update_tab_title(web_view, title))
            web_view.urlChanged.connect(lambda qurl: self.update_tab_url(web_view, qurl))
            web_view.loadStarted.connect(lambda: self.page_load_started(web_view))
            web_view.loadFinished.connect(lambda: self.page_load_finished(web_view))
            
            # Tab ekle
            if url == "about:blank":
                title = "Yeni Sekme"
            else:
                title = "Yükleniyor..."
                
            index = self.tab_widget.addTab(web_view, title)
            self.tabs[index] = tab_info
            self.tab_widget.setCurrentIndex(index)
            
            # URL'ye git
            if url != "about:blank":
                web_view.setUrl(QUrl(url))
            
            return index
        
        def close_tab(self, index: int):
            """Sekme kapat"""
            if self.tab_widget.count() <= 1:
                # Son sekme kapatılırsa yeni bir tane oluştur
                self.new_tab()
            
            # Tab bilgisini sil
            if index in self.tabs:
                del self.tabs[index]
            
            # Tab'ı kapat
            self.tab_widget.removeTab(index)
            
            # Index'leri yeniden düzenle
            new_tabs = {}
            for i, tab_index in enumerate(sorted(self.tabs.keys())):
                if tab_index > index:
                    new_tabs[tab_index - 1] = self.tabs[tab_index]
                else:
                    new_tabs[tab_index] = self.tabs[tab_index]
            self.tabs = new_tabs
        
        def close_current_tab(self):
            """Mevcut sekmeyi kapat"""
            current_index = self.tab_widget.currentIndex()
            if current_index != -1:
                self.close_tab(current_index)
        
        def tab_changed(self, index: int):
            """Sekme değişti"""
            self.current_tab_index = index
            if index != -1 and index in self.tabs:
                tab_info = self.tabs[index]
                web_view = self.tab_widget.widget(index)
                
                if isinstance(web_view, QWebEngineView):
                    current_url = web_view.url().toString()
                    self.address_bar.setText(current_url)
                    
                    # Navigasyon butonları
                    self.back_btn.setEnabled(web_view.history().canGoBack())
                    self.forward_btn.setEnabled(web_view.history().canGoForward())
                    
                    # Yer imi durumu
                    self.update_bookmark_button(current_url)
        
        def navigate_to_url(self):
            """Adres çubuğundaki URL'ye git"""
            url_text = self.address_bar.text().strip()
            if not url_text:
                return
            
            # URL formatını düzelt
            if not url_text.startswith(('http://', 'https://', 'file://')):
                if '.' in url_text and ' ' not in url_text:
                    url_text = 'https://' + url_text
                else:
                    # Arama yap
                    url_text = f'https://www.google.com/search?q={url_text}'
            
            current_widget = self.tab_widget.currentWidget()
            if isinstance(current_widget, QWebEngineView):
                current_widget.setUrl(QUrl(url_text))
        
        def go_back(self):
            """Geri git"""
            current_widget = self.tab_widget.currentWidget()
            if isinstance(current_widget, QWebEngineView):
                current_widget.back()
        
        def go_forward(self):
            """İleri git"""
            current_widget = self.tab_widget.currentWidget()
            if isinstance(current_widget, QWebEngineView):
                current_widget.forward()
        
        def reload_page(self):
            """Sayfayı yenile"""
            current_widget = self.tab_widget.currentWidget()
            if isinstance(current_widget, QWebEngineView):
                current_widget.reload()
        
        def go_home(self):
            """Ana sayfaya git"""
            home_url = "https://www.google.com"
            current_widget = self.tab_widget.currentWidget()
            if isinstance(current_widget, QWebEngineView):
                current_widget.setUrl(QUrl(home_url))
        
        def update_tab_title(self, web_view, title):
            """Sekme başlığını güncelle"""
            for index in range(self.tab_widget.count()):
                if self.tab_widget.widget(index) == web_view:
                    self.tab_widget.setTabText(index, title[:30] + "..." if len(title) > 30 else title)
                    if index in self.tabs:
                        self.tabs[index].title = title
                    break
        
        def update_tab_url(self, web_view, qurl):
            """Sekme URL'sini güncelle"""
            url = qurl.toString()
            for index in range(self.tab_widget.count()):
                if self.tab_widget.widget(index) == web_view:
                    if index in self.tabs:
                        self.tabs[index].url = url
                        self.tabs[index].add_to_history(url)
                    
                    # Adres çubuğunu güncelle (sadece aktif sekme için)
                    if index == self.tab_widget.currentIndex():
                        self.address_bar.setText(url)
                        self.update_bookmark_button(url)
                    break
        
        def page_load_started(self, web_view):
            """Sayfa yüklenmeye başladı"""
            for index in range(self.tab_widget.count()):
                if self.tab_widget.widget(index) == web_view:
                    if index in self.tabs:
                        self.tabs[index].is_loading = True
                    break
            
            self.status_bar.showMessage("Yükleniyor...")
        
        def page_load_finished(self, web_view):
            """Sayfa yüklenme bitti"""
            for index in range(self.tab_widget.count()):
                if self.tab_widget.widget(index) == web_view:
                    if index in self.tabs:
                        self.tabs[index].is_loading = False
                    break
            
            self.status_bar.showMessage("Hazır")
            
            # Navigasyon butonlarını güncelle
            if web_view == self.tab_widget.currentWidget():
                self.back_btn.setEnabled(web_view.history().canGoBack())
                self.forward_btn.setEnabled(web_view.history().canGoForward())
        
        def add_current_bookmark(self):
            """Mevcut sayfayı yer imine ekle"""
            current_widget = self.tab_widget.currentWidget()
            if isinstance(current_widget, QWebEngineView):
                url = current_widget.url().toString()
                title = current_widget.title() or "Başlıksız"
                
                if not self.bookmark_manager.is_bookmarked(url):
                    self.bookmark_manager.add_bookmark(title, url)
                    self.update_bookmark_button(url)
                    self.update_bookmark_menu()
                    self.status_bar.showMessage("Yer imine eklendi", 2000)
        
        def toggle_bookmark(self):
            """Yer imi ekle/çıkar"""
            current_widget = self.tab_widget.currentWidget()
            if isinstance(current_widget, QWebEngineView):
                url = current_widget.url().toString()
                
                if self.bookmark_manager.is_bookmarked(url):
                    self.bookmark_manager.remove_bookmark(url)
                    self.status_bar.showMessage("Yer iminden çıkarıldı", 2000)
                else:
                    title = current_widget.title() or "Başlıksız"
                    self.bookmark_manager.add_bookmark(title, url)
                    self.status_bar.showMessage("Yer imine eklendi", 2000)
                
                self.update_bookmark_button(url)
                self.update_bookmark_menu()
        
        def update_bookmark_button(self, url: str):
            """Yer imi butonunu güncelle"""
            if self.bookmark_manager.is_bookmarked(url):
                self.bookmark_btn.setText("★")
                self.bookmark_btn.setToolTip("Yer İminden Çıkar")
            else:
                self.bookmark_btn.setText("⭐")
                self.bookmark_btn.setToolTip("Yer İmine Ekle")
        
        def update_bookmark_menu(self):
            """Yer imi menüsünü güncelle"""
            # Bu fonksiyon menü güncelleme için kullanılacak
            pass
        
        def manage_bookmarks(self):
            """Yer imi yönetim penceresi"""
            dialog = QDialog(self)
            dialog.setWindowTitle("Yer İmi Yöneticisi")
            dialog.setGeometry(200, 200, 600, 400)
            
            layout = QVBoxLayout()
            
            # Yer imi listesi
            bookmark_list = QListWidget()
            for bookmark in self.bookmark_manager.get_bookmarks():
                item = QListWidgetItem(f"{bookmark['title']} - {bookmark['url']}")
                item.setData(Qt.ItemDataRole.UserRole, bookmark['url'])
                bookmark_list.addItem(item)
            
            layout.addWidget(bookmark_list)
            
            # Butonlar
            button_layout = QHBoxLayout()
            
            open_btn = QPushButton("Aç")
            open_btn.clicked.connect(lambda: self.open_bookmark_from_dialog(bookmark_list, dialog))
            button_layout.addWidget(open_btn)
            
            delete_btn = QPushButton("Sil")
            delete_btn.clicked.connect(lambda: self.delete_bookmark_from_dialog(bookmark_list))
            button_layout.addWidget(delete_btn)
            
            close_btn = QPushButton("Kapat")
            close_btn.clicked.connect(dialog.close)
            button_layout.addWidget(close_btn)
            
            layout.addLayout(button_layout)
            dialog.setLayout(layout)
            dialog.exec()
        
        def open_bookmark_from_dialog(self, bookmark_list, dialog):
            """Dialog'dan yer imi aç"""
            current_item = bookmark_list.currentItem()
            if current_item:
                url = current_item.data(Qt.ItemDataRole.UserRole)
                self.new_tab(url)
                dialog.close()
        
        def delete_bookmark_from_dialog(self, bookmark_list):
            """Dialog'dan yer imi sil"""
            current_item = bookmark_list.currentItem()
            if current_item:
                url = current_item.data(Qt.ItemDataRole.UserRole)
                self.bookmark_manager.remove_bookmark(url)
                bookmark_list.takeItem(bookmark_list.row(current_item))
        
        def next_tab(self):
            """Sonraki sekme"""
            current = self.tab_widget.currentIndex()
            count = self.tab_widget.count()
            next_index = (current + 1) % count
            self.tab_widget.setCurrentIndex(next_index)
        
        def prev_tab(self):
            """Önceki sekme"""
            current = self.tab_widget.currentIndex()
            count = self.tab_widget.count()
            prev_index = (current - 1) % count
            self.tab_widget.setCurrentIndex(prev_index)
        
        def reopen_closed_tab(self):
            """Kapatılmış sekmeyi yeniden aç"""
            # Basit implementasyon - geliştirilecek
            self.new_tab("https://www.google.com")
        
        def new_window(self):
            """Yeni pencere aç"""
            new_browser = BrowserWindow(self.kernel)
            new_browser.show()
        
        def toggle_fullscreen(self):
            """Tam ekran değiştir"""
            if self.isFullScreen():
                self.showNormal()
            else:
                self.showFullScreen()

elif PYQT_AVAILABLE and not WEBENGINE_AVAILABLE:
    class SimpleBrowserWindow(QMainWindow):
        """Basit tarayıcı penceresi (WebEngine olmadan)"""
        
        def __init__(self, kernel=None):
            super().__init__()
            self.kernel = kernel
            self.bookmark_manager = BookmarkManager()
            
            self.init_ui()
        
        def init_ui(self):
            """UI'yı başlat"""
            self.setWindowTitle("PyCloud Browser (Basit Mod)")
            self.setGeometry(100, 100, 800, 600)
            
            central_widget = QWidget()
            self.setCentralWidget(central_widget)
            
            layout = QVBoxLayout()
            
            # Adres çubuğu
            address_layout = QHBoxLayout()
            self.address_bar = QLineEdit()
            self.address_bar.setPlaceholderText("Dosya yolu girin (file:// veya http://)")
            self.address_bar.returnPressed.connect(self.load_content)
            
            load_btn = QPushButton("Git")
            load_btn.clicked.connect(self.load_content)
            
            address_layout.addWidget(self.address_bar)
            address_layout.addWidget(load_btn)
            layout.addLayout(address_layout)
            
            # İçerik alanı
            self.content_area = QTextEdit()
            self.content_area.setReadOnly(True)
            layout.addWidget(self.content_area)
            
            central_widget.setLayout(layout)
            
            # Başlangıç mesajı
            self.content_area.setText("""
PyCloud Browser - Basit Mod

WebEngine mevcut olmadığı için basit metin görüntüleme modu aktif.

Desteklenen özellikler:
- Yerel HTML dosyalarını görüntüleme (file://)
- Metin dosyalarını okuma
- Basit HTTP istekleri

Örnek kullanım:
- file:///Users/kullanici/dosya.html
- file:///home/kullanici/dosya.txt
- http://example.com (basit içerik)
            """)
        
        def load_content(self):
            """İçerik yükle"""
            url = self.address_bar.text().strip()
            if not url:
                return
            
            try:
                if url.startswith('file://'):
                    # Yerel dosya
                    file_path = url[7:]  # file:// kısmını çıkar
                    with open(file_path, 'r', encoding='utf-8') as f:
                        content = f.read()
                    
                    if file_path.endswith('.html'):
                        # HTML dosyası olarak göster
                        self.content_area.setHtml(content)
                    else:
                        # Düz metin olarak göster
                        self.content_area.setPlainText(content)
                
                elif url.startswith(('http://', 'https://')):
                    # Basit HTTP isteği
                    import urllib.request
                    with urllib.request.urlopen(url) as response:
                        content = response.read().decode('utf-8')
                    self.content_area.setPlainText(content[:5000] + "..." if len(content) > 5000 else content)
                
                else:
                    self.content_area.setPlainText("Desteklenmeyen URL formatı")
                    
            except Exception as e:
                self.content_area.setPlainText(f"Hata: {e}")

# Text-mode Browser (PyQt6 yoksa)
class TextBrowser:
    """Text-mode browser"""
    
    def __init__(self, kernel=None):
        self.kernel = kernel
        self.bookmark_manager = BookmarkManager()
    
    def run(self):
        """Browser'ı çalıştır"""
        print("PyCloud Browser v1.0 (Text Mode)")
        print("Komutlar: go <url>, bookmarks, add <title> <url>, quit")
        print()
        
        while True:
            try:
                command = input("browser> ").strip()
                if not command:
                    continue
                
                if not self.handle_command(command):
                    break
                    
            except KeyboardInterrupt:
                print("\nBrowser kapatılıyor...")
                break
            except EOFError:
                break
    
    def handle_command(self, command: str) -> bool:
        """Komut işle"""
        parts = command.split(maxsplit=1)
        if not parts:
            return True
        
        cmd = parts[0].lower()
        
        if cmd == 'quit' or cmd == 'exit':
            return False
        
        elif cmd == 'go':
            if len(parts) > 1:
                self.visit_url(parts[1])
            else:
                print("Kullanım: go <url>")
        
        elif cmd == 'bookmarks':
            self.show_bookmarks()
        
        elif cmd == 'add':
            if len(parts) > 1:
                bookmark_parts = parts[1].split(maxsplit=1)
                if len(bookmark_parts) >= 2:
                    title, url = bookmark_parts
                    self.bookmark_manager.add_bookmark(title, url)
                    print(f"Yer imi eklendi: {title}")
                else:
                    print("Kullanım: add <başlık> <url>")
            else:
                print("Kullanım: add <başlık> <url>")
        
        else:
            print(f"Bilinmeyen komut: {cmd}")
            print("Komutlar: go, bookmarks, add, quit")
        
        return True
    
    def visit_url(self, url: str):
        """URL'yi ziyaret et"""
        try:
            if url.startswith('file://'):
                file_path = url[7:]
                with open(file_path, 'r', encoding='utf-8') as f:
                    content = f.read()
                print(f"\n=== {file_path} ===")
                print(content[:500] + "..." if len(content) > 500 else content)
            
            elif url.startswith(('http://', 'https://')):
                import urllib.request
                with urllib.request.urlopen(url) as response:
                    content = response.read().decode('utf-8')
                print(f"\n=== {url} ===")
                print(content[:500] + "..." if len(content) > 500 else content)
            
            else:
                print("Desteklenmeyen URL formatı")
                
        except Exception as e:
            print(f"Hata: {e}")
    
    def show_bookmarks(self):
        """Yer imlerini göster"""
        bookmarks = self.bookmark_manager.get_bookmarks()
        if bookmarks:
            print("\nYer İmleri:")
            for i, bookmark in enumerate(bookmarks, 1):
                print(f"{i}. {bookmark['title']} - {bookmark['url']}")
        else:
            print("Yer imi yok")

# Ana fonksiyonlar
def create_browser_app(kernel=None):
    """Browser uygulaması oluştur"""
    if PYQT_AVAILABLE:
        app = QApplication.instance()
        if app is None:
            app = QApplication(sys.argv)
        
        if WEBENGINE_AVAILABLE:
            browser = BrowserWindow(kernel)
        else:
            browser = SimpleBrowserWindow(kernel)
        
        browser.show()
        return browser
    else:
        return TextBrowser(kernel)

def run_browser(kernel=None):
    """Browser'ı çalıştır"""
    if PYQT_AVAILABLE:
        # QApplication oluştur
        app = QApplication.instance()
        if app is None:
            app = QApplication(sys.argv)
        
        if WEBENGINE_AVAILABLE:
            browser = BrowserWindow(kernel)
        else:
            browser = SimpleBrowserWindow(kernel)
        
        # ✅ ÇÖZÜM: Command line argumentlarını parse et
        import argparse
        parser = argparse.ArgumentParser(description='Cloud Browser')
        parser.add_argument('--open-file', dest='open_file', help='Açılacak dosya yolu')
        parser.add_argument('--url', dest='url', help='Açılacak URL')
        parser.add_argument('files', nargs='*', help='Açılacak dosyalar/URL\'ler')
        
        # sys.argv'yi parse et
        try:
            args, unknown = parser.parse_known_args()
            
            # Dosya açma parametresi varsa
            if args.open_file:
                file_path = Path(args.open_file).absolute()
                file_url = f"file://{file_path}"
                print(f"🚀 Browser dosya açıyor: {file_url}")
                if hasattr(browser, 'new_tab'):
                    browser.new_tab(file_url)
                else:
                    browser.address_bar.setText(file_url)
                    browser.load_content()
            
            # URL parametresi varsa
            elif args.url:
                print(f"🚀 Browser URL açıyor: {args.url}")
                if hasattr(browser, 'new_tab'):
                    browser.new_tab(args.url)
                else:
                    browser.address_bar.setText(args.url)
                    browser.load_content()
            
            # Doğrudan dosya/URL listesi varsa
            elif args.files:
                for item in args.files:
                    if Path(item).exists():
                        # Dosya ise file:// protokolü ekle
                        file_path = Path(item).absolute()
                        file_url = f"file://{file_path}"
                        print(f"🚀 Browser dosya açıyor: {file_url}")
                        if hasattr(browser, 'new_tab'):
                            browser.new_tab(file_url)
                        else:
                            browser.address_bar.setText(file_url)
                            browser.load_content()
                            break  # Basit mode'da sadece ilkini aç
                    elif item.startswith(('http://', 'https://', 'file://')):
                        # URL ise doğrudan aç
                        print(f"🚀 Browser URL açıyor: {item}")
                        if hasattr(browser, 'new_tab'):
                            browser.new_tab(item)
                        else:
                            browser.address_bar.setText(item)
                            browser.load_content()
                            break  # Basit mode'da sadece ilkini aç
                        
        except Exception as e:
            print(f"⚠️ Browser argument parsing error: {e}")
            # Argumentlar parse edilemezse normal başlat
        
        browser.show()
        return browser
    else:
        browser = TextBrowser(kernel)
        
        # Text mode için de dosya/URL açma desteği
        import argparse
        parser = argparse.ArgumentParser(description='Cloud Browser (Text Mode)')
        parser.add_argument('--open-file', dest='open_file', help='Açılacak dosya yolu')
        parser.add_argument('--url', dest='url', help='Açılacak URL')
        parser.add_argument('files', nargs='*', help='Açılacak dosyalar/URL\'ler')
        
        try:
            args, unknown = parser.parse_known_args()
            
            if args.open_file and Path(args.open_file).exists():
                file_path = Path(args.open_file).absolute()
                file_url = f"file://{file_path}"
                print(f"🚀 Browser (Text) dosya açıyor: {file_url}")
                browser.visit_url(file_url)
            elif args.url:
                print(f"🚀 Browser (Text) URL açıyor: {args.url}")
                browser.visit_url(args.url)
            elif args.files:
                for item in args.files:
                    if Path(item).exists():
                        file_path = Path(item).absolute()
                        file_url = f"file://{file_path}"
                        print(f"🚀 Browser (Text) dosya açıyor: {file_url}")
                        browser.visit_url(file_url)
                        break
                    elif item.startswith(('http://', 'https://', 'file://')):
                        print(f"🚀 Browser (Text) URL açıyor: {item}")
                        browser.visit_url(item)
                        break
        except Exception as e:
            print(f"⚠️ Browser (Text) argument parsing error: {e}")
        
        browser.run()
        return None

if __name__ == "__main__":
    run_browser() 


#!/usr/bin/env python3
"""
new_command.py - Yeni Uygulama Oluşturma Komutu

Bu modül clapp için yeni uygulama şablonları oluşturur.
"""

import os
import shutil
import json
from pathlib import Path
from package_runner import get_supported_languages

def create_new_app(language: str, app_name: str, target_dir: str | None = None) -> tuple[bool, str]:
    """
    Yeni bir clapp uygulaması oluşturur
    
    Args:
        language: Programlama dili
        app_name: Uygulama adı
        target_dir: Hedef dizin (opsiyonel)
        
    Returns:
        (success, message)
    """
    try:
        # Dil kontrolü
        if language.lower() not in get_supported_languages():
            supported = ', '.join(get_supported_languages())
            return False, f"Desteklenmeyen dil: {language}. Desteklenen diller: {supported}"
        
        # Uygulama adı kontrolü
        if not app_name or not app_name.strip():
            return False, "Uygulama adı boş olamaz"
        
        app_name = app_name.strip().lower()
        
        # Geçersiz karakter kontrolü
        invalid_chars = ['/', '\\', ':', '*', '?', '"', '<', '>', '|']
        for char in invalid_chars:
            if char in app_name:
                return False, f"Uygulama adında geçersiz karakter: {char}"
        
        # Hedef dizin belirleme
        if target_dir:
            target_path = Path(target_dir) / app_name
        else:
            target_path = Path.cwd() / app_name
        
        # Dizin zaten var mı kontrol et
        if target_path.exists():
            return False, f"Dizin zaten mevcut: {target_path}"
        
        # Şablon dizini
        template_dir = Path(__file__).parent / "templates" / language.lower()
        
        if not template_dir.exists():
            return False, f"Şablon bulunamadı: {language}"
        
        # Dizini oluştur
        target_path.mkdir(parents=True, exist_ok=True)
        
        # Şablon dosyalarını kopyala
        for item in template_dir.iterdir():
            if item.is_file():
                # Dosya adını uygulama adına göre güncelle
                if item.name == "manifest.json":
                    update_manifest(item, target_path / item.name, app_name, language)
                else:
                    shutil.copy2(item, target_path / item.name)
            elif item.is_dir():
                # Alt dizinleri kopyala
                shutil.copytree(item, target_path / item.name)
        
        # Başarı mesajı
        success_message = f"""
✅ Yeni {language} uygulaması oluşturuldu!

📁 Dizin: {target_path}
🚀 Çalıştırmak için:
   cd {app_name}
   clapp validate .
   clapp install .
   clapp run {app_name}

📖 Geliştirme için docs/developer_guide.md dosyasını inceleyin.
        """.strip()
        
        return True, success_message
        
    except Exception as e:
        return False, f"Uygulama oluşturma hatası: {str(e)}"

def update_manifest(template_path: Path, target_path: Path, app_name: str, language: str):
    """
    Manifest dosyasını uygulama adına göre günceller
    
    Args:
        template_path: Şablon manifest dosyası
        target_path: Hedef manifest dosyası
        app_name: Uygulama adı
        language: Programlama dili
    """
    try:
        # Şablon manifest'i oku
        with open(template_path, 'r', encoding='utf-8') as f:
            manifest = json.load(f)
        
        # Uygulama adını güncelle
        manifest['name'] = app_name
        
        # Açıklamayı güncelle
        if 'description' in manifest:
            manifest['description'] = manifest['description'].replace(
                'hello-' + language.lower(), app_name
            )
        
        # Hedef dosyaya yaz
        with open(target_path, 'w', encoding='utf-8') as f:
            json.dump(manifest, f, indent=2, ensure_ascii=False)
            
    except Exception as e:
        print(f"⚠️  Manifest güncelleme hatası: {e}")

def list_available_templates() -> str:
    """
    Mevcut şablonları listeler
    
    Returns:
        Formatlanmış şablon listesi
    """
    templates_dir = Path(__file__).parent / "templates"
    
    if not templates_dir.exists():
        return "❌ Şablon dizini bulunamadı"
    
    result = "📋 Mevcut Şablonlar:\n"
    result += "=" * 30 + "\n\n"
    
    for template in sorted(templates_dir.iterdir()):
        if template.is_dir():
            # Şablon bilgilerini oku
            manifest_path = template / "manifest.json"
            if manifest_path.exists():
                try:
                    with open(manifest_path, 'r', encoding='utf-8') as f:
                        manifest = json.load(f)
                    
                    name = manifest.get('name', template.name)
                    description = manifest.get('description', 'Açıklama yok')
                    language = manifest.get('language', 'unknown')
                    
                    result += f"🌐 {language.upper()}\n"
                    result += f"   📝 {description}\n"
                    result += f"   📁 Şablon: {template.name}\n\n"
                    
                except Exception:
                    result += f"🌐 {template.name.upper()}\n"
                    result += f"   📁 Şablon: {template.name}\n\n"
    
    result += "💡 Kullanım: clapp new <dil> <uygulama-adı>"
    
    return result

def handle_new_command(args) -> tuple[bool, str]:
    """
    new komutunu işler
    
    Args:
        args: Argümanlar
        
    Returns:
        (success, message)
    """
    if args.list:
        return True, list_available_templates()
    
    if not args.language or not args.app_name:
        return False, "Dil ve uygulama adı gerekli. Örnek: clapp new python my-app"
    
    return create_new_app(args.language, args.app_name, args.target_dir) 


"""
Cloud Task Manager - Modern UI Widget'ları
macOS Activity Monitor tarzı modern bileşenler
"""

import sys
from typing import List, Optional
from collections import deque

try:
    from PyQt6.QtWidgets import *
    from PyQt6.QtCore import *
    from PyQt6.QtGui import *
except ImportError:
    print("PyQt6 not available for widgets")
    sys.exit(1)

class ModernTabWidget(QTabWidget):
    """Modern sekmeli widget"""
    
    def __init__(self):
        super().__init__()
        self.setTabPosition(QTabWidget.TabPosition.North)
        self.setMovable(False)
        self.setTabsClosable(False)
        
        # Modern stil
        self.setStyleSheet("""
            QTabWidget::pane {
                border: 1px solid #dee2e6;
                border-radius: 8px;
                background-color: #ffffff;
                margin-top: -1px;
            }
            
            QTabBar::tab {
                background-color: #f8f9fa;
                border: 1px solid #dee2e6;
                padding: 12px 20px;
                margin-right: 2px;
                border-top-left-radius: 8px;
                border-top-right-radius: 8px;
                font-weight: 600;
                font-size: 14px;
            }
            
            QTabBar::tab:selected {
                background-color: #ffffff;
                border-bottom-color: #ffffff;
                color: #007bff;
            }
            
            QTabBar::tab:hover {
                background-color: #e9ecef;
            }
        """)

class SystemOverviewWidget(QWidget):
    """Sistem özeti widget'ı"""
    
    def __init__(self):
        super().__init__()
        self.setFixedHeight(120)
        self.setup_ui()
        
    def setup_ui(self):
        """UI kurulumu"""
        layout = QHBoxLayout(self)
        layout.setContentsMargins(20, 15, 20, 15)
        layout.setSpacing(30)
        
        # CPU
        self.cpu_widget = self.create_metric_widget("CPU", "0%", "#007bff")
        layout.addWidget(self.cpu_widget)
        
        # Memory
        self.memory_widget = self.create_metric_widget("Memory", "0%", "#28a745")
        layout.addWidget(self.memory_widget)
        
        # Disk
        self.disk_widget = self.create_metric_widget("Disk", "0%", "#ffc107")
        layout.addWidget(self.disk_widget)
        
        # Network
        self.network_widget = self.create_metric_widget("Network", "0 MB/s", "#dc3545")
        layout.addWidget(self.network_widget)
        
        # Processes
        self.processes_widget = self.create_metric_widget("Processes", "0", "#6f42c1")
        layout.addWidget(self.processes_widget)
        
        # Modern stil
        self.setStyleSheet("""
            SystemOverviewWidget {
                background-color: #ffffff;
                border: 1px solid #dee2e6;
                border-radius: 12px;
            }
        """)
    
    def create_metric_widget(self, title: str, value: str, color: str) -> QWidget:
        """Metrik widget'ı oluştur"""
        widget = QWidget()
        layout = QVBoxLayout(widget)
        layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        layout.setSpacing(5)
        
        # Başlık
        title_label = QLabel(title)
        title_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        title_label.setStyleSheet(f"""
            font-size: 12px;
            font-weight: 600;
            color: #6c757d;
            margin-bottom: 5px;
        """)
        layout.addWidget(title_label)
        
        # Değer
        value_label = QLabel(value)
        value_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        value_label.setStyleSheet(f"""
            font-size: 24px;
            font-weight: 700;
            color: {color};
        """)
        layout.addWidget(value_label)
        
        # Progress bar
        progress = QProgressBar()
        progress.setRange(0, 100)
        progress.setValue(0)
        progress.setTextVisible(False)
        progress.setFixedHeight(6)
        progress.setStyleSheet(f"""
            QProgressBar {{
                border: none;
                border-radius: 3px;
                background-color: #e9ecef;
            }}
            
            QProgressBar::chunk {{
                background-color: {color};
                border-radius: 3px;
            }}
        """)
        layout.addWidget(progress)
        
        # Widget'a referansları sakla
        widget.title_label = title_label
        widget.value_label = value_label
        widget.progress = progress
        
        return widget
    
    def update_stats(self, stats):
        """İstatistikleri güncelle"""
        # CPU
        self.cpu_widget.value_label.setText(f"{stats.cpu_percent:.1f}%")
        self.cpu_widget.progress.setValue(int(stats.cpu_percent))
        
        # Memory
        self.memory_widget.value_label.setText(f"{stats.memory_percent:.1f}%")
        self.memory_widget.progress.setValue(int(stats.memory_percent))
        
        # Disk
        self.disk_widget.value_label.setText(f"{stats.disk_percent:.1f}%")
        self.disk_widget.progress.setValue(int(stats.disk_percent))
        
        # Network
        self.network_widget.value_label.setText(f"{stats.network_mb:.1f} MB/s")
        network_percent = min(100, (stats.network_mb / 10) * 100)  # 10 MB/s = 100%
        self.network_widget.progress.setValue(int(network_percent))
        
        # Processes
        self.processes_widget.value_label.setText(str(stats.process_count))
        process_percent = min(100, (stats.process_count / 500) * 100)  # 500 process = 100%
        self.processes_widget.progress.setValue(int(process_percent))

class ProcessTableWidget(QTableWidget):
    """Modern işlem tablosu"""
    
    def __init__(self, table_type: str = "processes"):
        super().__init__()
        self.table_type = table_type
        self.setup_table()
    
    def setup_table(self):
        """Tablo kurulumu"""
        self.setAlternatingRowColors(True)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setSortingEnabled(True)
        self.setShowGrid(False)
        
        # Header ayarları
        header = self.horizontalHeader()
        header.setDefaultAlignment(Qt.AlignmentFlag.AlignLeft)
        header.setHighlightSections(False)
        
        # Vertical header gizle
        self.verticalHeader().setVisible(False)
        
        # Modern stil
        self.setStyleSheet("""
            QTableWidget {
                border: 1px solid #dee2e6;
                border-radius: 8px;
                background-color: #ffffff;
                gridline-color: #f8f9fa;
                selection-background-color: #007bff;
                selection-color: white;
                font-size: 13px;
            }
            
            QTableWidget::item {
                padding: 8px;
                border-bottom: 1px solid #f8f9fa;
            }
            
            QTableWidget::item:selected {
                background-color: #007bff;
                color: white;
            }
            
            QTableWidget::item:hover {
                background-color: #f8f9fa;
            }
            
            QHeaderView::section {
                background-color: #f8f9fa;
                border: none;
                border-bottom: 2px solid #dee2e6;
                padding: 12px 8px;
                font-weight: 600;
                font-size: 12px;
                color: #495057;
            }
            
            QHeaderView::section:hover {
                background-color: #e9ecef;
            }
        """)

class ThreadTableWidget(QTableWidget):
    """Modern thread tablosu"""
    
    def __init__(self):
        super().__init__()
        self.setup_table()
    
    def setup_table(self):
        """Tablo kurulumu"""
        self.setAlternatingRowColors(True)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setSortingEnabled(True)
        self.setShowGrid(False)
        
        # Header ayarları
        header = self.horizontalHeader()
        header.setDefaultAlignment(Qt.AlignmentFlag.AlignLeft)
        header.setHighlightSections(False)
        
        # Vertical header gizle
        self.verticalHeader().setVisible(False)
        
        # Thread tablosu için özel stil
        self.setStyleSheet("""
            QTableWidget {
                border: 1px solid #dee2e6;
                border-radius: 8px;
                background-color: #ffffff;
                gridline-color: #f8f9fa;
                selection-background-color: #6f42c1;
                selection-color: white;
                font-size: 13px;
            }
            
            QTableWidget::item {
                padding: 8px;
                border-bottom: 1px solid #f8f9fa;
            }
            
            QTableWidget::item:selected {
                background-color: #6f42c1;
                color: white;
            }
            
            QTableWidget::item:hover {
                background-color: #f8f9fa;
            }
            
            QHeaderView::section {
                background-color: #f8f9fa;
                border: none;
                border-bottom: 2px solid #dee2e6;
                padding: 12px 8px;
                font-weight: 600;
                font-size: 12px;
                color: #495057;
            }
        """)

class ResourceGraphWidget(QWidget):
    """Kaynak kullanımı grafiği"""
    
    def __init__(self, title: str, unit: str, max_value: float = 100):
        super().__init__()
        self.title = title
        self.unit = unit
        self.max_value = max_value
        self.data_points = deque(maxlen=60)  # Son 60 veri noktası
        self.setMinimumSize(300, 200)
        
        # Başlangıç verileri
        for _ in range(60):
            self.data_points.append(0)
    
    def add_data_point(self, value: float):
        """Yeni veri noktası ekle"""
        self.data_points.append(value)
        self.update()  # Widget'ı yeniden çiz
    
    def paintEvent(self, event):
        """Grafik çizimi"""
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        
        # Arka plan
        painter.fillRect(self.rect(), QColor("#ffffff"))
        
        # Kenarlık
        painter.setPen(QPen(QColor("#dee2e6"), 1))
        painter.drawRoundedRect(self.rect().adjusted(0, 0, -1, -1), 8, 8)
        
        # Başlık
        painter.setPen(QColor("#495057"))
        painter.setFont(QFont("Arial", 12, QFont.Weight.Bold))
        title_rect = QRect(10, 10, self.width() - 20, 20)
        painter.drawText(title_rect, Qt.AlignmentFlag.AlignLeft, self.title)
        
        # Grafik alanı
        graph_rect = QRect(20, 40, self.width() - 40, self.height() - 70)
        
        # Grid çizgileri
        painter.setPen(QPen(QColor("#f8f9fa"), 1))
        for i in range(5):
            y = graph_rect.top() + (graph_rect.height() * i / 4)
            painter.drawLine(graph_rect.left(), y, graph_rect.right(), y)
        
        # Veri çizgisi
        if len(self.data_points) > 1:
            painter.setPen(QPen(QColor("#007bff"), 2))
            
            points = []
            for i, value in enumerate(self.data_points):
                x = graph_rect.left() + (graph_rect.width() * i / (len(self.data_points) - 1))
                y = graph_rect.bottom() - (graph_rect.height() * value / self.max_value)
                points.append(QPointF(x, y))
            
            # Çizgiyi çiz
            for i in range(len(points) - 1):
                painter.drawLine(points[i], points[i + 1])
        
        # Değer etiketi
        if self.data_points:
            current_value = self.data_points[-1]
            painter.setPen(QColor("#495057"))
            painter.setFont(QFont("Arial", 10))
            value_text = f"{current_value:.1f} {self.unit}"
            value_rect = QRect(10, self.height() - 25, self.width() - 20, 20)
            painter.drawText(value_rect, Qt.AlignmentFlag.AlignRight, value_text)

class ModernProgressBar(QProgressBar):
    """Modern progress bar"""
    
    def __init__(self, color: str = "#007bff"):
        super().__init__()
        self.color = color
        self.setTextVisible(False)
        self.setFixedHeight(8)
        
        self.setStyleSheet(f"""
            QProgressBar {{
                border: none;
                border-radius: 4px;
                background-color: #e9ecef;
            }}
            
            QProgressBar::chunk {{
                background-color: {color};
                border-radius: 4px;
            }}
        """)

class ModernButton(QPushButton):
    """Modern buton"""
    
    def __init__(self, text: str, button_type: str = "primary"):
        super().__init__(text)
        
        colors = {
            "primary": "#007bff",
            "success": "#28a745",
            "danger": "#dc3545",
            "warning": "#ffc107",
            "secondary": "#6c757d"
        }
        
        color = colors.get(button_type, "#007bff")
        
        self.setStyleSheet(f"""
            QPushButton {{
                background-color: {color};
                color: white;
                border: none;
                padding: 10px 20px;
                border-radius: 6px;
                font-weight: 600;
                font-size: 14px;
            }}
            
            QPushButton:hover {{
                background-color: {self._darken_color(color)};
            }}
            
            QPushButton:pressed {{
                background-color: {self._darken_color(color, 0.2)};
            }}
            
            QPushButton:disabled {{
                background-color: #e9ecef;
                color: #6c757d;
            }}
        """)
    
    def _darken_color(self, color: str, factor: float = 0.1) -> str:
        """Rengi koyulaştır"""
        # Basit renk koyulaştırma
        if color == "#007bff":
            return "#0056b3"
        elif color == "#28a745":
            return "#1e7e34"
        elif color == "#dc3545":
            return "#c82333"
        elif color == "#ffc107":
            return "#e0a800"
        elif color == "#6c757d":
            return "#545b62"
        return color

class ModernSearchBox(QLineEdit):
    """Modern arama kutusu"""
    
    def __init__(self, placeholder: str = "Search..."):
        super().__init__()
        self.setPlaceholderText(placeholder)
        
        self.setStyleSheet("""
            QLineEdit {
                border: 2px solid #e9ecef;
                border-radius: 8px;
                padding: 8px 12px;
                font-size: 14px;
                background-color: #ffffff;
            }
            
            QLineEdit:focus {
                border-color: #007bff;
                outline: none;
            }
            
            QLineEdit:hover {
                border-color: #dee2e6;
            }
        """)

class StatusIndicator(QWidget):
    """Durum göstergesi"""
    
    def __init__(self, status: str = "unknown"):
        super().__init__()
        self.status = status
        self.setFixedSize(12, 12)
    
    def set_status(self, status: str):
        """Durumu değiştir"""
        self.status = status
        self.update()
    
    def paintEvent(self, event):
        """Durum göstergesi çizimi"""
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        
        colors = {
            "running": "#28a745",
            "stopped": "#dc3545",
            "suspended": "#ffc107",
            "unknown": "#6c757d"
        }
        
        color = QColor(colors.get(self.status, "#6c757d"))
        painter.setBrush(QBrush(color))
        painter.setPen(Qt.PenStyle.NoPen)
        painter.drawEllipse(self.rect()) 


"""Low-level infrastructure to find modules.

This builds on fscache.py; find_sources.py builds on top of this.
"""

from __future__ import annotations

import ast
import collections
import functools
import os
import re
import subprocess
import sys
from enum import Enum, unique
from typing import Final, Optional, Union
from typing_extensions import TypeAlias as _TypeAlias

from pathspec import PathSpec
from pathspec.patterns.gitwildmatch import GitWildMatchPatternError

from mypy import pyinfo
from mypy.errors import CompileError
from mypy.fscache import FileSystemCache
from mypy.nodes import MypyFile
from mypy.options import Options
from mypy.stubinfo import stub_distribution_name
from mypy.util import os_path_join


# Paths to be searched in find_module().
class SearchPaths:
    def __init__(
        self,
        python_path: tuple[str, ...],
        mypy_path: tuple[str, ...],
        package_path: tuple[str, ...],
        typeshed_path: tuple[str, ...],
    ) -> None:
        # where user code is found
        self.python_path = tuple(map(os.path.abspath, python_path))
        # from $MYPYPATH or config variable
        self.mypy_path = tuple(map(os.path.abspath, mypy_path))
        # from get_site_packages_dirs()
        self.package_path = tuple(map(os.path.abspath, package_path))
        # paths in typeshed
        self.typeshed_path = tuple(map(os.path.abspath, typeshed_path))

    def asdict(self) -> dict[str, tuple[str, ...]]:
        return {
            "python_path": self.python_path,
            "mypy_path": self.mypy_path,
            "package_path": self.package_path,
            "typeshed_path": self.typeshed_path,
        }


# Package dirs are a two-tuple of path to search and whether to verify the module
OnePackageDir = tuple[str, bool]
PackageDirs = list[OnePackageDir]

# Minimum and maximum Python versions for modules in stdlib as (major, minor)
StdlibVersions: _TypeAlias = dict[str, tuple[tuple[int, int], Optional[tuple[int, int]]]]

PYTHON_EXTENSIONS: Final = [".pyi", ".py"]


# TODO: Consider adding more reasons here?
# E.g. if we deduce a module would likely be found if the user were
# to set the --namespace-packages flag.
@unique
class ModuleNotFoundReason(Enum):
    # The module was not found: we found neither stubs nor a plausible code
    # implementation (with or without a py.typed file).
    NOT_FOUND = 0

    # The implementation for this module plausibly exists (e.g. we
    # found a matching folder or *.py file), but either the parent package
    # did not contain a py.typed file or we were unable to find a
    # corresponding *-stubs package.
    FOUND_WITHOUT_TYPE_HINTS = 1

    # The module was not found in the current working directory, but
    # was able to be found in the parent directory.
    WRONG_WORKING_DIRECTORY = 2

    # Stub PyPI package (typically types-pkgname) known to exist but not installed.
    APPROVED_STUBS_NOT_INSTALLED = 3

    def error_message_templates(self, daemon: bool) -> tuple[str, list[str]]:
        doc_link = "See https://mypy.readthedocs.io/en/stable/running_mypy.html#missing-imports"
        if self is ModuleNotFoundReason.NOT_FOUND:
            msg = 'Cannot find implementation or library stub for module named "{module}"'
            notes = [doc_link]
        elif self is ModuleNotFoundReason.WRONG_WORKING_DIRECTORY:
            msg = 'Cannot find implementation or library stub for module named "{module}"'
            notes = [
                "You may be running mypy in a subpackage, mypy should be run on the package root"
            ]
        elif self is ModuleNotFoundReason.FOUND_WITHOUT_TYPE_HINTS:
            msg = (
                'Skipping analyzing "{module}": module is installed, but missing library stubs '
                "or py.typed marker"
            )
            notes = [doc_link]
        elif self is ModuleNotFoundReason.APPROVED_STUBS_NOT_INSTALLED:
            msg = 'Library stubs not installed for "{module}"'
            notes = ['Hint: "python3 -m pip install {stub_dist}"']
            if not daemon:
                notes.append(
                    '(or run "mypy --install-types" to install all missing stub packages)'
                )
            notes.append(doc_link)
        else:
            assert False
        return msg, notes


# If we found the module, returns the path to the module as a str.
# Otherwise, returns the reason why the module wasn't found.
ModuleSearchResult = Union[str, ModuleNotFoundReason]


class BuildSource:
    """A single source file."""

    def __init__(
        self,
        path: str | None,
        module: str | None,
        text: str | None = None,
        base_dir: str | None = None,
        followed: bool = False,
    ) -> None:
        self.path = path  # File where it's found (e.g. 'xxx/yyy/foo/bar.py')
        self.module = module or "__main__"  # Module name (e.g. 'foo.bar')
        self.text = text  # Source code, if initially supplied, else None
        self.base_dir = base_dir  # Directory where the package is rooted (e.g. 'xxx/yyy')
        self.followed = followed  # Was this found by following imports?

    def __repr__(self) -> str:
        return (
            "BuildSource(path={!r}, module={!r}, has_text={}, base_dir={!r}, followed={})".format(
                self.path, self.module, self.text is not None, self.base_dir, self.followed
            )
        )


class BuildSourceSet:
    """Helper to efficiently test a file's membership in a set of build sources."""

    def __init__(self, sources: list[BuildSource]) -> None:
        self.source_text_present = False
        self.source_modules: dict[str, str] = {}
        self.source_paths: set[str] = set()

        for source in sources:
            if source.text is not None:
                self.source_text_present = True
            if source.path:
                self.source_paths.add(source.path)
            if source.module:
                self.source_modules[source.module] = source.path or ""

    def is_source(self, file: MypyFile) -> bool:
        return (
            (file.path and file.path in self.source_paths)
            or file._fullname in self.source_modules
            or self.source_text_present
        )


class FindModuleCache:
    """Module finder with integrated cache.

    Module locations and some intermediate results are cached internally
    and can be cleared with the clear() method.

    All file system accesses are performed through a FileSystemCache,
    which is not ever cleared by this class. If necessary it must be
    cleared by client code.
    """

    def __init__(
        self,
        search_paths: SearchPaths,
        fscache: FileSystemCache | None,
        options: Options | None,
        stdlib_py_versions: StdlibVersions | None = None,
        source_set: BuildSourceSet | None = None,
    ) -> None:
        self.search_paths = search_paths
        self.source_set = source_set
        self.fscache = fscache or FileSystemCache()
        # Cache for get_toplevel_possibilities:
        # search_paths -> (toplevel_id -> list(package_dirs))
        self.initial_components: dict[tuple[str, ...], dict[str, list[str]]] = {}
        # Cache find_module: id -> result
        self.results: dict[str, ModuleSearchResult] = {}
        self.ns_ancestors: dict[str, str] = {}
        self.options = options
        custom_typeshed_dir = None
        if options:
            custom_typeshed_dir = options.custom_typeshed_dir
        self.stdlib_py_versions = stdlib_py_versions or load_stdlib_py_versions(
            custom_typeshed_dir
        )

    def clear(self) -> None:
        self.results.clear()
        self.initial_components.clear()
        self.ns_ancestors.clear()

    def find_module_via_source_set(self, id: str) -> ModuleSearchResult | None:
        """Fast path to find modules by looking through the input sources

        This is only used when --fast-module-lookup is passed on the command line."""
        if not self.source_set:
            return None

        p = self.source_set.source_modules.get(id, None)
        if p and self.fscache.isfile(p):
            # We need to make sure we still have __init__.py all the way up
            # otherwise we might have false positives compared to slow path
            # in case of deletion of init files, which is covered by some tests.
            # TODO: are there some combination of flags in which this check should be skipped?
            d = os.path.dirname(p)
            for _ in range(id.count(".")):
                if not any(
                    self.fscache.isfile(os_path_join(d, "__init__" + x)) for x in PYTHON_EXTENSIONS
                ):
                    return None
                d = os.path.dirname(d)
            return p

        idx = id.rfind(".")
        if idx != -1:
            # When we're looking for foo.bar.baz and can't find a matching module
            # in the source set, look up for a foo.bar module.
            parent = self.find_module_via_source_set(id[:idx])
            if parent is None or not isinstance(parent, str):
                return None

            basename, ext = os.path.splitext(parent)
            if not any(parent.endswith("__init__" + x) for x in PYTHON_EXTENSIONS) and (
                ext in PYTHON_EXTENSIONS and not self.fscache.isdir(basename)
            ):
                # If we do find such a *module* (and crucially, we don't want a package,
                # hence the filtering out of __init__ files, and checking for the presence
                # of a folder with a matching name), then we can be pretty confident that
                # 'baz' will either be a top-level variable in foo.bar, or will not exist.
                #
                # Either way, spelunking in other search paths for another 'foo.bar.baz'
                # module should be avoided because:
                #  1. in the unlikely event that one were found, it's highly likely that
                #     it would be unrelated to the source being typechecked and therefore
                #     more likely to lead to erroneous results
                #  2. as described in _find_module, in some cases the search itself could
                #  potentially waste significant amounts of time
                return ModuleNotFoundReason.NOT_FOUND
        return None

    def find_lib_path_dirs(self, id: str, lib_path: tuple[str, ...]) -> PackageDirs:
        """Find which elements of a lib_path have the directory a module needs to exist."""
        components = id.split(".")
        dir_chain = os.sep.join(components[:-1])  # e.g., 'foo/bar'

        dirs = []
        for pathitem in self.get_toplevel_possibilities(lib_path, components[0]):
            # e.g., '/usr/lib/python3.4/foo/bar'
            if dir_chain:
                dir = os_path_join(pathitem, dir_chain)
            else:
                dir = pathitem
            if self.fscache.isdir(dir):
                dirs.append((dir, True))
        return dirs

    def get_toplevel_possibilities(self, lib_path: tuple[str, ...], id: str) -> list[str]:
        """Find which elements of lib_path could contain a particular top-level module.

        In practice, almost all modules can be routed to the correct entry in
        lib_path by looking at just the first component of the module name.

        We take advantage of this by enumerating the contents of all of the
        directories on the lib_path and building a map of which entries in
        the lib_path could contain each potential top-level module that appears.
        """

        if lib_path in self.initial_components:
            return self.initial_components[lib_path].get(id, [])

        # Enumerate all the files in the directories on lib_path and produce the map
        components: dict[str, list[str]] = {}
        for dir in lib_path:
            try:
                contents = self.fscache.listdir(dir)
            except OSError:
                contents = []
            # False positives are fine for correctness here, since we will check
            # precisely later, so we only look at the root of every filename without
            # any concern for the exact details.
            for name in contents:
                name = os.path.splitext(name)[0]
                components.setdefault(name, []).append(dir)

        self.initial_components[lib_path] = components
        return components.get(id, [])

    def find_module(self, id: str, *, fast_path: bool = False) -> ModuleSearchResult:
        """Return the path of the module source file or why it wasn't found.

        If fast_path is True, prioritize performance over generating detailed
        error descriptions.
        """
        if id not in self.results:
            top_level = id.partition(".")[0]
            use_typeshed = True
            if id in self.stdlib_py_versions:
                use_typeshed = self._typeshed_has_version(id)
            elif top_level in self.stdlib_py_versions:
                use_typeshed = self._typeshed_has_version(top_level)
            result, should_cache = self._find_module(id, use_typeshed)
            if should_cache:
                if (
                    not (
                        fast_path or (self.options is not None and self.options.fast_module_lookup)
                    )
                    and result is ModuleNotFoundReason.NOT_FOUND
                    and self._can_find_module_in_parent_dir(id)
                ):
                    self.results[id] = ModuleNotFoundReason.WRONG_WORKING_DIRECTORY
                else:
                    self.results[id] = result
                return self.results[id]
            else:
                return result
        return self.results[id]

    def _typeshed_has_version(self, module: str) -> bool:
        if not self.options:
            return True
        version = typeshed_py_version(self.options)
        min_version, max_version = self.stdlib_py_versions[module]
        return version >= min_version and (max_version is None or version <= max_version)

    def _find_module_non_stub_helper(
        self, id: str, pkg_dir: str
    ) -> OnePackageDir | ModuleNotFoundReason:
        plausible_match = False
        dir_path = pkg_dir
        components = id.split(".")
        for index, component in enumerate(components):
            dir_path = os_path_join(dir_path, component)
            if self.fscache.isfile(os_path_join(dir_path, "py.typed")):
                return os.path.join(pkg_dir, *components[:-1]), index == 0
            elif not plausible_match and (
                self.fscache.isdir(dir_path) or self.fscache.isfile(dir_path + ".py")
            ):
                plausible_match = True
            # If this is not a directory then we can't traverse further into it
            if not self.fscache.isdir(dir_path):
                break
        if plausible_match:
            if self.options:
                module_specific_options = self.options.clone_for_module(id)
                if module_specific_options.follow_untyped_imports:
                    return os.path.join(pkg_dir, *components[:-1]), False
            return ModuleNotFoundReason.FOUND_WITHOUT_TYPE_HINTS
        else:
            return ModuleNotFoundReason.NOT_FOUND

    def _update_ns_ancestors(self, components: list[str], match: tuple[str, bool]) -> None:
        path, verify = match
        for i in range(1, len(components)):
            pkg_id = ".".join(components[:-i])
            if pkg_id not in self.ns_ancestors and self.fscache.isdir(path):
                self.ns_ancestors[pkg_id] = path
            path = os.path.dirname(path)

    def _can_find_module_in_parent_dir(self, id: str) -> bool:
        """Test if a module can be found by checking the parent directories
        of the current working directory.
        """
        working_dir = os.getcwd()
        parent_search = FindModuleCache(
            SearchPaths((), (), (), ()),
            self.fscache,
            self.options,
            stdlib_py_versions=self.stdlib_py_versions,
        )
        while any(is_init_file(file) for file in os.listdir(working_dir)):
            working_dir = os.path.dirname(working_dir)
            parent_search.search_paths = SearchPaths((working_dir,), (), (), ())
            if not isinstance(parent_search._find_module(id, False)[0], ModuleNotFoundReason):
                return True
        return False

    def _find_module(self, id: str, use_typeshed: bool) -> tuple[ModuleSearchResult, bool]:
        """Try to find a module in all available sources.

        Returns:
            ``(result, can_be_cached)`` pair.
        """
        fscache = self.fscache

        # Fast path for any modules in the current source set.
        # This is particularly important when there are a large number of search
        # paths which share the first (few) component(s) due to the use of namespace
        # packages, for instance:
        # foo/
        #    company/
        #        __init__.py
        #        foo/
        # bar/
        #    company/
        #        __init__.py
        #        bar/
        # baz/
        #    company/
        #        __init__.py
        #        baz/
        #
        # mypy gets [foo/company/foo, bar/company/bar, baz/company/baz, ...] as input
        # and computes [foo, bar, baz, ...] as the module search path.
        #
        # This would result in O(n) search for every import of company.*, leading to
        # O(n**2) behavior in load_graph as such imports are unsurprisingly present
        # at least once, and usually many more times than that, in each and every file
        # being parsed.
        #
        # Thankfully, such cases are efficiently handled by looking up the module path
        # via BuildSourceSet.
        p = (
            self.find_module_via_source_set(id)
            if (self.options is not None and self.options.fast_module_lookup)
            else None
        )
        if p:
            return p, True

        # If we're looking for a module like 'foo.bar.baz', it's likely that most of the
        # many elements of lib_path don't even have a subdirectory 'foo/bar'.  Discover
        # that only once and cache it for when we look for modules like 'foo.bar.blah'
        # that will require the same subdirectory.
        components = id.split(".")
        dir_chain = os.sep.join(components[:-1])  # e.g., 'foo/bar'

        # We have two sets of folders so that we collect *all* stubs folders and
        # put them in the front of the search path
        third_party_inline_dirs: PackageDirs = []
        third_party_stubs_dirs: PackageDirs = []
        found_possible_third_party_missing_type_hints = False
        # Third-party stub/typed packages
        candidate_package_dirs = {
            package_dir[0]
            for component in (components[0], components[0] + "-stubs")
            for package_dir in self.find_lib_path_dirs(component, self.search_paths.package_path)
        }
        # Caching FOUND_WITHOUT_TYPE_HINTS is not always safe. That causes issues with
        # typed subpackages in namespace packages.
        can_cache_any_result = True
        for pkg_dir in self.search_paths.package_path:
            if pkg_dir not in candidate_package_dirs:
                continue
            stub_name = components[0] + "-stubs"
            stub_dir = os_path_join(pkg_dir, stub_name)
            if fscache.isdir(stub_dir):
                stub_typed_file = os_path_join(stub_dir, "py.typed")
                stub_components = [stub_name] + components[1:]
                path = os.path.join(pkg_dir, *stub_components[:-1])
                if fscache.isdir(path):
                    if fscache.isfile(stub_typed_file):
                        # Stub packages can have a py.typed file, which must include
                        # 'partial\n' to make the package partial
                        # Partial here means that mypy should look at the runtime
                        # package if installed.
                        if fscache.read(stub_typed_file).decode().strip() == "partial":
                            runtime_path = os_path_join(pkg_dir, dir_chain)
                            third_party_inline_dirs.append((runtime_path, True))
                            # if the package is partial, we don't verify the module, as
                            # the partial stub package may not have a __init__.pyi
                            third_party_stubs_dirs.append((path, False))
                        else:
                            # handle the edge case where people put a py.typed file
                            # in a stub package, but it isn't partial
                            third_party_stubs_dirs.append((path, True))
                    else:
                        third_party_stubs_dirs.append((path, True))
            non_stub_match = self._find_module_non_stub_helper(id, pkg_dir)
            if isinstance(non_stub_match, ModuleNotFoundReason):
                if non_stub_match is ModuleNotFoundReason.FOUND_WITHOUT_TYPE_HINTS:
                    found_possible_third_party_missing_type_hints = True
                    can_cache_any_result = False
            else:
                third_party_inline_dirs.append(non_stub_match)
                self._update_ns_ancestors(components, non_stub_match)

        if self.options and self.options.use_builtins_fixtures:
            # Everything should be in fixtures.
            third_party_inline_dirs.clear()
            third_party_stubs_dirs.clear()
            found_possible_third_party_missing_type_hints = False
        python_mypy_path = self.search_paths.mypy_path + self.search_paths.python_path
        candidate_base_dirs = self.find_lib_path_dirs(id, python_mypy_path)
        if use_typeshed:
            # Search for stdlib stubs in typeshed before installed
            # stubs to avoid picking up backports (dataclasses, for
            # example) when the library is included in stdlib.
            candidate_base_dirs += self.find_lib_path_dirs(id, self.search_paths.typeshed_path)
        candidate_base_dirs += third_party_stubs_dirs + third_party_inline_dirs

        # If we're looking for a module like 'foo.bar.baz', then candidate_base_dirs now
        # contains just the subdirectories 'foo/bar' that actually exist under the
        # elements of lib_path.  This is probably much shorter than lib_path itself.
        # Now just look for 'baz.pyi', 'baz/__init__.py', etc., inside those directories.
        seplast = os.sep + components[-1]  # so e.g. '/baz'
        sepinit = os.sep + "__init__"
        near_misses = []  # Collect near misses for namespace mode (see below).
        for base_dir, verify in candidate_base_dirs:
            base_path = base_dir + seplast  # so e.g. '/usr/lib/python3.4/foo/bar/baz'
            has_init = False
            dir_prefix = base_dir
            for _ in range(len(components) - 1):
                dir_prefix = os.path.dirname(dir_prefix)

            # Stubs-only packages always take precedence over py.typed packages
            path_stubs = f"{base_path}-stubs{sepinit}.pyi"
            if fscache.isfile_case(path_stubs, dir_prefix):
                if verify and not verify_module(fscache, id, path_stubs, dir_prefix):
                    near_misses.append((path_stubs, dir_prefix))
                else:
                    return path_stubs, True

            # Prefer package over module, i.e. baz/__init__.py* over baz.py*.
            for extension in PYTHON_EXTENSIONS:
                path = base_path + sepinit + extension
                if fscache.isfile_case(path, dir_prefix):
                    has_init = True
                    if verify and not verify_module(fscache, id, path, dir_prefix):
                        near_misses.append((path, dir_prefix))
                        continue
                    return path, True

            # In namespace mode, register a potential namespace package
            if self.options and self.options.namespace_packages:
                if (
                    not has_init
                    and fscache.exists_case(base_path, dir_prefix)
                    and not fscache.isfile_case(base_path, dir_prefix)
                ):
                    near_misses.append((base_path, dir_prefix))

            # No package, look for module.
            for extension in PYTHON_EXTENSIONS:
                path = base_path + extension
                if fscache.isfile_case(path, dir_prefix):
                    if verify and not verify_module(fscache, id, path, dir_prefix):
                        near_misses.append((path, dir_prefix))
                        continue
                    return path, True

        # In namespace mode, re-check those entries that had 'verify'.
        # Assume search path entries xxx, yyy and zzz, and we're
        # looking for foo.bar.baz.  Suppose near_misses has:
        #
        # - xxx/foo/bar/baz.py
        # - yyy/foo/bar/baz/__init__.py
        # - zzz/foo/bar/baz.pyi
        #
        # If any of the foo directories has __init__.py[i], it wins.
        # Else, we look for foo/bar/__init__.py[i], etc.  If there are
        # none, the first hit wins.  Note that this does not take into
        # account whether the lowest-level module is a file (baz.py),
        # a package (baz/__init__.py), or a stub file (baz.pyi) -- for
        # these the first one encountered along the search path wins.
        #
        # The helper function highest_init_level() returns an int that
        # indicates the highest level at which a __init__.py[i] file
        # is found; if no __init__ was found it returns 0, if we find
        # only foo/bar/__init__.py it returns 1, and if we have
        # foo/__init__.py it returns 2 (regardless of what's in
        # foo/bar).  It doesn't look higher than that.
        if self.options and self.options.namespace_packages and near_misses:
            levels = [
                highest_init_level(fscache, id, path, dir_prefix)
                for path, dir_prefix in near_misses
            ]
            index = levels.index(max(levels))
            return near_misses[index][0], True

        # Finally, we may be asked to produce an ancestor for an
        # installed package with a py.typed marker that is a
        # subpackage of a namespace package.  We only fess up to these
        # if we would otherwise return "not found".
        ancestor = self.ns_ancestors.get(id)
        if ancestor is not None:
            return ancestor, True

        approved_dist_name = stub_distribution_name(id)
        if approved_dist_name:
            if len(components) == 1:
                return ModuleNotFoundReason.APPROVED_STUBS_NOT_INSTALLED, True
            # If we're a missing submodule of an already installed approved stubs, we don't want to
            # error with APPROVED_STUBS_NOT_INSTALLED, but rather want to return NOT_FOUND.
            for i in range(1, len(components)):
                parent_id = ".".join(components[:i])
                if stub_distribution_name(parent_id) == approved_dist_name:
                    break
            else:
                return ModuleNotFoundReason.APPROVED_STUBS_NOT_INSTALLED, True
            if self.find_module(parent_id) is ModuleNotFoundReason.APPROVED_STUBS_NOT_INSTALLED:
                return ModuleNotFoundReason.APPROVED_STUBS_NOT_INSTALLED, True
            return ModuleNotFoundReason.NOT_FOUND, True

        if found_possible_third_party_missing_type_hints:
            return ModuleNotFoundReason.FOUND_WITHOUT_TYPE_HINTS, can_cache_any_result
        return ModuleNotFoundReason.NOT_FOUND, True

    def find_modules_recursive(self, module: str) -> list[BuildSource]:
        module_path = self.find_module(module, fast_path=True)
        if isinstance(module_path, ModuleNotFoundReason):
            return []
        sources = [BuildSource(module_path, module, None)]

        package_path = None
        if is_init_file(module_path):
            package_path = os.path.dirname(module_path)
        elif self.fscache.isdir(module_path):
            package_path = module_path
        if package_path is None:
            return sources

        # This logic closely mirrors that in find_sources. One small but important difference is
        # that we do not sort names with keyfunc. The recursive call to find_modules_recursive
        # calls find_module, which will handle the preference between packages, pyi and py.
        # Another difference is it doesn't handle nested search paths / package roots.

        seen: set[str] = set()
        names = sorted(self.fscache.listdir(package_path))
        for name in names:
            # Skip certain names altogether
            if name in ("__pycache__", "site-packages", "node_modules") or name.startswith("."):
                continue
            subpath = os_path_join(package_path, name)

            if self.options and matches_exclude(
                subpath, self.options.exclude, self.fscache, self.options.verbosity >= 2
            ):
                continue
            if (
                self.options
                and self.options.exclude_gitignore
                and matches_gitignore(subpath, self.fscache, self.options.verbosity >= 2)
            ):
                continue

            if self.fscache.isdir(subpath):
                # Only recurse into packages
                if (self.options and self.options.namespace_packages) or (
                    self.fscache.isfile(os_path_join(subpath, "__init__.py"))
                    or self.fscache.isfile(os_path_join(subpath, "__init__.pyi"))
                ):
                    seen.add(name)
                    sources.extend(self.find_modules_recursive(module + "." + name))
            else:
                stem, suffix = os.path.splitext(name)
                if stem == "__init__":
                    continue
                if stem not in seen and "." not in stem and suffix in PYTHON_EXTENSIONS:
                    # (If we sorted names by keyfunc) we could probably just make the BuildSource
                    # ourselves, but this ensures compatibility with find_module / the cache
                    seen.add(stem)
                    sources.extend(self.find_modules_recursive(module + "." + stem))
        return sources


def matches_exclude(
    subpath: str, excludes: list[str], fscache: FileSystemCache, verbose: bool
) -> bool:
    if not excludes:
        return False
    subpath_str = os.path.relpath(subpath).replace(os.sep, "/")
    if fscache.isdir(subpath):
        subpath_str += "/"
    for exclude in excludes:
        try:
            if re.search(exclude, subpath_str):
                if verbose:
                    print(
                        f"TRACE: Excluding {subpath_str} (matches pattern {exclude})",
                        file=sys.stderr,
                    )
                return True
        except re.error as e:
            print(
                f"error: The exclude {exclude} is an invalid regular expression, because: {e}"
                + (
                    "\n(Hint: use / as a path separator, even if you're on Windows!)"
                    if "\\" in exclude
                    else ""
                )
                + "\nFor more information on Python's flavor of regex, see:"
                + " https://docs.python.org/3/library/re.html",
                file=sys.stderr,
            )
            sys.exit(2)
    return False


def matches_gitignore(subpath: str, fscache: FileSystemCache, verbose: bool) -> bool:
    dir, _ = os.path.split(subpath)
    for gi_path, gi_spec in find_gitignores(dir):
        relative_path = os.path.relpath(subpath, gi_path)
        if fscache.isdir(relative_path):
            relative_path = relative_path + "/"
        if gi_spec.match_file(relative_path):
            if verbose:
                print(
                    f"TRACE: Excluding {relative_path} (matches .gitignore) in {gi_path}",
                    file=sys.stderr,
                )
            return True
    return False


@functools.lru_cache
def find_gitignores(dir: str) -> list[tuple[str, PathSpec]]:
    parent_dir = os.path.dirname(dir)
    if parent_dir == dir:
        parent_gitignores = []
    else:
        parent_gitignores = find_gitignores(parent_dir)

    gitignore = os.path.join(dir, ".gitignore")
    if os.path.isfile(gitignore):
        with open(gitignore) as f:
            lines = f.readlines()
        try:
            return parent_gitignores + [(dir, PathSpec.from_lines("gitwildmatch", lines))]
        except GitWildMatchPatternError:
            print(f"error: could not parse {gitignore}", file=sys.stderr)
            return parent_gitignores
    return parent_gitignores


def is_init_file(path: str) -> bool:
    return os.path.basename(path) in ("__init__.py", "__init__.pyi")


def verify_module(fscache: FileSystemCache, id: str, path: str, prefix: str) -> bool:
    """Check that all packages containing id have a __init__ file."""
    if is_init_file(path):
        path = os.path.dirname(path)
    for i in range(id.count(".")):
        path = os.path.dirname(path)
        if not any(
            fscache.isfile_case(os_path_join(path, f"__init__{extension}"), prefix)
            for extension in PYTHON_EXTENSIONS
        ):
            return False
    return True


def highest_init_level(fscache: FileSystemCache, id: str, path: str, prefix: str) -> int:
    """Compute the highest level where an __init__ file is found."""
    if is_init_file(path):
        path = os.path.dirname(path)
    level = 0
    for i in range(id.count(".")):
        path = os.path.dirname(path)
        if any(
            fscache.isfile_case(os_path_join(path, f"__init__{extension}"), prefix)
            for extension in PYTHON_EXTENSIONS
        ):
            level = i + 1
    return level


def mypy_path() -> list[str]:
    path_env = os.getenv("MYPYPATH")
    if not path_env:
        return []
    return path_env.split(os.pathsep)


def default_lib_path(
    data_dir: str, pyversion: tuple[int, int], custom_typeshed_dir: str | None
) -> list[str]:
    """Return default standard library search paths. Guaranteed to be normalised."""

    data_dir = os.path.abspath(data_dir)
    path: list[str] = []

    if custom_typeshed_dir:
        custom_typeshed_dir = os.path.abspath(custom_typeshed_dir)
        typeshed_dir = os.path.join(custom_typeshed_dir, "stdlib")
        mypy_extensions_dir = os.path.join(custom_typeshed_dir, "stubs", "mypy-extensions")
        versions_file = os.path.join(typeshed_dir, "VERSIONS")
        if not os.path.isdir(typeshed_dir) or not os.path.isfile(versions_file):
            print(
                "error: --custom-typeshed-dir does not point to a valid typeshed ({})".format(
                    custom_typeshed_dir
                ),
                file=sys.stderr,
            )
            sys.exit(2)
    else:
        auto = os.path.join(data_dir, "stubs-auto")
        if os.path.isdir(auto):
            data_dir = auto
        typeshed_dir = os.path.join(data_dir, "typeshed", "stdlib")
        mypy_extensions_dir = os.path.join(data_dir, "typeshed", "stubs", "mypy-extensions")
    path.append(typeshed_dir)

    # Get mypy-extensions stubs from typeshed, since we treat it as an
    # "internal" library, similar to typing and typing-extensions.
    path.append(mypy_extensions_dir)

    # Add fallback path that can be used if we have a broken installation.
    if sys.platform != "win32":
        path.append("/usr/local/lib/mypy")
    if not path:
        print(
            "Could not resolve typeshed subdirectories. Your mypy install is broken.\n"
            "Python executable is located at {}.\nMypy located at {}".format(
                sys.executable, data_dir
            ),
            file=sys.stderr,
        )
        sys.exit(1)
    return path


@functools.cache
def get_search_dirs(python_executable: str | None) -> tuple[list[str], list[str]]:
    """Find package directories for given python. Guaranteed to return absolute paths.

    This runs a subprocess call, which generates a list of the directories in sys.path.
    To avoid repeatedly calling a subprocess (which can be slow!) we
    lru_cache the results.
    """

    if python_executable is None:
        return ([], [])
    elif python_executable == sys.executable:
        # Use running Python's package dirs
        sys_path, site_packages = pyinfo.getsearchdirs()
    else:
        # Use subprocess to get the package directory of given Python
        # executable
        env = {**dict(os.environ), "PYTHONSAFEPATH": "1"}
        try:
            sys_path, site_packages = ast.literal_eval(
                subprocess.check_output(
                    [python_executable, pyinfo.__file__, "getsearchdirs"],
                    env=env,
                    stderr=subprocess.PIPE,
                ).decode()
            )
        except subprocess.CalledProcessError as err:
            print(err.stderr)
            print(err.stdout)
            raise
        except OSError as err:
            assert err.errno is not None
            reason = os.strerror(err.errno)
            raise CompileError(
                [f"mypy: Invalid python executable '{python_executable}': {reason}"]
            ) from err
    return sys_path, site_packages


def compute_search_paths(
    sources: list[BuildSource], options: Options, data_dir: str, alt_lib_path: str | None = None
) -> SearchPaths:
    """Compute the search paths as specified in PEP 561.

    There are the following 4 members created:
    - User code (from `sources`)
    - MYPYPATH (set either via config or environment variable)
    - installed package directories (which will later be split into stub-only and inline)
    - typeshed
    """
    # Determine the default module search path.
    lib_path = collections.deque(
        default_lib_path(
            data_dir, options.python_version, custom_typeshed_dir=options.custom_typeshed_dir
        )
    )

    if options.use_builtins_fixtures:
        # Use stub builtins (to speed up test cases and to make them easier to
        # debug).  This is a test-only feature, so assume our files are laid out
        # as in the source tree.
        # We also need to allow overriding where to look for it. Argh.
        root_dir = os.getenv("MYPY_TEST_PREFIX", None)
        if not root_dir:
            root_dir = os.path.dirname(os.path.dirname(__file__))
        root_dir = os.path.abspath(root_dir)
        lib_path.appendleft(os.path.join(root_dir, "test-data", "unit", "lib-stub"))
    # alt_lib_path is used by some tests to bypass the normal lib_path mechanics.
    # If we don't have one, grab directories of source files.
    python_path: list[str] = []
    if not alt_lib_path:
        for source in sources:
            # Include directory of the program file in the module search path.
            if source.base_dir:
                dir = source.base_dir
                if dir not in python_path:
                    python_path.append(dir)

        # Do this even if running as a file, for sanity (mainly because with
        # multiple builds, there could be a mix of files/modules, so its easier
        # to just define the semantics that we always add the current director
        # to the lib_path
        # TODO: Don't do this in some cases; for motivation see see
        # https://github.com/python/mypy/issues/4195#issuecomment-341915031
        if options.bazel:
            dir = "."
        else:
            dir = os.getcwd()
        if dir not in lib_path:
            python_path.insert(0, dir)

    # Start with a MYPYPATH environment variable at the front of the mypy_path, if defined.
    mypypath = mypy_path()

    # Add a config-defined mypy path.
    mypypath.extend(options.mypy_path)

    # If provided, insert the caller-supplied extra module path to the
    # beginning (highest priority) of the search path.
    if alt_lib_path:
        mypypath.insert(0, alt_lib_path)

    sys_path, site_packages = get_search_dirs(options.python_executable)
    # We only use site packages for this check
    for site in site_packages:
        assert site not in lib_path
        if (
            site in mypypath
            or any(p.startswith(site + os.path.sep) for p in mypypath)
            or (os.path.altsep and any(p.startswith(site + os.path.altsep) for p in mypypath))
        ):
            print(f"{site} is in the MYPYPATH. Please remove it.", file=sys.stderr)
            print(
                "See https://mypy.readthedocs.io/en/stable/running_mypy.html"
                "#how-mypy-handles-imports for more info",
                file=sys.stderr,
            )
            sys.exit(1)

    return SearchPaths(
        python_path=tuple(reversed(python_path)),
        mypy_path=tuple(mypypath),
        package_path=tuple(sys_path + site_packages),
        typeshed_path=tuple(lib_path),
    )


def load_stdlib_py_versions(custom_typeshed_dir: str | None) -> StdlibVersions:
    """Return dict with minimum and maximum Python versions of stdlib modules.

    The contents look like
    {..., 'secrets': ((3, 6), None), 'symbol': ((2, 7), (3, 9)), ...}

    None means there is no maximum version.
    """
    typeshed_dir = custom_typeshed_dir or os_path_join(os.path.dirname(__file__), "typeshed")
    stdlib_dir = os_path_join(typeshed_dir, "stdlib")
    result = {}

    versions_path = os_path_join(stdlib_dir, "VERSIONS")
    assert os.path.isfile(versions_path), (custom_typeshed_dir, versions_path, __file__)
    with open(versions_path) as f:
        for line in f:
            line = line.split("#")[0].strip()
            if line == "":
                continue
            module, version_range = line.split(":")
            versions = version_range.split("-")
            min_version = parse_version(versions[0])
            max_version = (
                parse_version(versions[1]) if len(versions) >= 2 and versions[1].strip() else None
            )
            result[module] = min_version, max_version
    return result


def parse_version(version: str) -> tuple[int, int]:
    major, minor = version.strip().split(".")
    return int(major), int(minor)


def typeshed_py_version(options: Options) -> tuple[int, int]:
    """Return Python version used for checking whether module supports typeshed."""
    # Typeshed no longer covers Python 3.x versions before 3.9, so 3.9 is
    # the earliest we can support.
    return max(options.python_version, (3, 9))
